/*
  This file is a collection of globally accessible utility functions. They're intended to be used from anywhere you 
  might need them, which could be in various extension pages, or in content scripts, etc. This is design ed to be a
  small library, so don't include massive amounts of stuff. Think C standard library. A bit of text processing, etc.
  These are also all inserted into the global namespace, so no 'Evernote' object needs to be defined to use them.
*/

var GlobalUtils = {};
(function(){
  "use strict";

  GlobalUtils.NOTEBOOK_TYPE_PERSONAL = 1;
  GlobalUtils.NOTEBOOK_TYPE_LINKED = 2;
  GlobalUtils.NOTEBOOK_TYPE_BUSINESS = 3;

  var urlMatcher = /^(.*?):\/\/((www\.)?(.*?))(:\d+)?(\/.*?)(\?.*)?$/;

  // var BAD_FAV_ICON_URLS = {"http://localhost/favicon.ico": true};

  GlobalUtils.componentizeUrl = function(url) {
    var data = {
      protocol: null,
      domain: null,
      domainNoWww: null,
      port: null,
      path: null,
      queryString: null
    };
    var matches = urlMatcher.exec(url);
    data.protocol = matches[1];
    data.domain = matches[2];
    data.domainNoWww = matches[4];
    data.port = matches[5];
    data.path = matches[6];
    data.queryString = matches[7];
    return data;
  };

  GlobalUtils.localize = function(element) {
    var node = element.nodeName.toLowerCase();
    if (node == "input" || node == "textarea") {
      switch (element.type) {
        case "text":
          if (element.attributes["placeholder"]) {
            var localizedMessage = Browser.i18n.getMessage(element.attributes["placeholder"].value);
            if (localizedMessage) {
              element.placeholder = localizedMessage;
            }
          }
        case "textarea":
        case "button":
        case "submit":
        case "search":
          if (element.attributes["placeholder"]) {
            var localizedMessage = Browser.i18n.getMessage(element.attributes["placeholder"].value);
            if (localizedMessage) {
              element.placeholder = localizedMessage;
            }
          }
          if (element.attributes["message"]) {
            var localizedMessage = Browser.i18n.getMessage(element.attributes["message"].value);
            if (localizedMessage) {
              element.value = localizedMessage;
            }
          }
          break;
          // unlocalizable.
        case "checkbox":
        case "password":
        case "hidden":
        case "radio":
          break;

        default:
          throw new Error("We need to localize the value of input elements.");
      }
    }

    else if (element.attributes["message"]) {
      var localizedMessage = Browser.i18n.getMessage(element.attributes["message"].value);
      if (localizedMessage) {
        element.innerHTML = localizedMessage;
      }
    }

    if (element.title){
      var localizedTitle = Browser.i18n.getMessage(element.title);
      if (localizedTitle) {
        element.title = localizedTitle;
      }
    }

    for (var i = 0; i < element.children.length; i++) {
      GlobalUtils.localize(element.children[i]);
    }
  };

  GlobalUtils.getQueryParams = function(url) {
    var data = GlobalUtils.componentizeUrl(url);
    var queryString = data.queryString;
    var params = {};
    if (!queryString) {
      return params;
    }
    queryString = queryString.substr(1); // Don't want the question mark.
    queryString = queryString.split("#")[0]; // Get rid of any fragment identifier.
    var pairs = queryString.split("&");
    var i;
    for (i = 0; i < pairs.length; i++) {
      var item = pairs[i].split("=");
      if (item[1]) {
        item[1] = item[1].replace(/\+/g, " ");
      }
      params[item[0].toLowerCase()] = item[1];
    }
    return params;
  };

  GlobalUtils.escapeXML = function(str) {
    return str.replace(/&|<|>|"|'/g, function(match) {
      if (match == "&") {
        return "&amp;";
      } else if (match == "<") {
        return "&lt;";
      } else if (match == ">") {
        return "&gt;";
      } else if (match == "\"") {
        return "&quot;";
      } else if (match == "'") {
        return "&apos;";
      }
    });
  };

  GlobalUtils.unescapeXML = function(str) {
    return str.replace(/&(amp|lt|gt|quot|apos);/g, function(match) {
      if (match == "&amp;") {
        return "&";
      } else if (match == "&lt;") {
        return "<";
      } else if (match == "&gt;") {
        return ">";
      } else if (match == "&quot;") {
        return "\"";
      } else if (match == "&apos;") {
        return "'";
      }
    });
  };

  // generates the URI for the web and desktop client versions of normal notes, linked notes, and biz notes
  GlobalUtils.getNoteURI = function(baseUrl, message, userId, token) {
    return baseUrl + "/SetAuthToken.action?auth=" + encodeURIComponent(token)
      + "&targetUrl=" + encodeURIComponent("/shard/" + message.shardId
      + "/nl/" + userId + "/" + message.noteGuid + "/");
  };

  GlobalUtils.buildGoogleRegEx = function() {
    // http://www.google.com/supported_domains
    var googleCountryDomains = [".com",".ad",".ae",".com.af",".com.ag",".com.ai",".am",".co.ao",".com.ar",".as",".at",
      ".com.au",".az",".ba",".com.bd",".be",".bf",".bg",".com.bh",".bi",".bj",".com.bn",".com.bo",".com.br",".bs",
      ".co.bw",".by",".com.bz",".ca",".cd",".cf",".cg",".ch",".ci",".co.ck",".cl",".cm",".cn",".com.co",".co.cr",
      ".com.cu",".cv",".com.cy",".cz",".de",".dj",".dk",".dm",".com.do",".dz",".com.ec",".ee",".com.eg",".es",".com.et",
      ".fi",".com.fj",".fm",".fr",".ga",".ge",".gg",".com.gh",".com.gi",".gl",".gm",".gp",".gr",".com.gt",".gy",
      ".com.hk",".hn",".hr",".ht",".hu",".co.id",".ie",".co.il",".im",".co.in",".iq",".is",".it",".je",".com.jm",".jo",
      ".co.jp",".co.ke",".com.kh",".ki",".kg",".co.kr",".com.kw",".kz",".la",".com.lb",".li",".lk",".co.ls",".lt",".lu",
      ".lv",".com.ly",".co.ma",".md",".me",".mg",".mk",".ml",".mn",".ms",".com.mt",".mu",".mv",".mw",".com.mx",".com.my",
      ".co.mz",".com.na",".com.nf",".com.ng",".com.ni",".ne",".nl",".no",".com.np",".nr",".nu",".co.nz",".com.om",
      ".com.pa",".com.pe",".com.ph",".com.pk",".pl",".pn",".com.pr",".ps",".pt",".com.py",".com.qa",".ro",".ru",".rw",
      ".com.sa",".com.sb",".sc",".se",".com.sg",".sh",".si",".sk",".com.sl",".sn",".so",".sm",".st",".com.sv",".td",
      ".tg",".co.th",".com.tj",".tk",".tl",".tm",".tn",".to",".com.tr",".tt",".com.tw",".co.tz",".com.ua",".co.ug",
      ".co.uk",".com.uy",".co.uz",".com.vc",".co.ve",".vg",".co.vi",".com.vn",".vu",".ws",".rs",".co.za",".co.zm",
      ".co.zw",".cat"];
    for (var i = 0; i < googleCountryDomains.length; i++) {
      googleCountryDomains[i] = googleCountryDomains[i].replace(/\./g, "\\.");
    }
    return new RegExp("^https?://www\.google(" + googleCountryDomains.join("|") + ")/", "i");
  };

  GlobalUtils.removePunctuation = function(text) {
    var dashes = ["-", "_", "\u2013", "\u2014", "\u00B7"];
    var brackets = ["\\(", "\\)", "\\[", "\\]", "\\{", "\\}", "\u300A", "\u300B",
      "\uFF08", "\uFF09", "\u3010", "\u3011", "\u300C", "\u300D", "\u00BB"];
    var punc = ["\\.", "!", "\:", "\;", "\"", "\'", ",", "\\?", "\u3002",
      "\u3001", "\uFF01", "\uFF0C", "\uFF1A", "\u2026", "\u201C", "\u201D"];
    var misc = ["@", "#", "\\$", "%", "\\^", "&", "\\*", "\\+", "=", "`", "~",
      "/", "\\\\", "\\|", ">", "<", "\u25CF" ];

    var regex = new RegExp(dashes.join("|") + "|" + brackets.join("|") + "|" + punc.join("|") + "|" + misc.join("|"), "g");
    return text.replace(regex, " ");
  };

  GlobalUtils.setupCache = function(obj, maxKeys, maxAge) {
    return function() {
      function cullCache(userId, key, callback) {
        // remove stale records
        IDB.deleteGroup(userId, key, "timestamp", IDBKeyRange.upperBound(new Date(new Date() - maxAge), true), callback, function(msg) {
          console.error(msg);
        });
      }

      obj.get = function(userId, type, key, callback) {
        cullCache(userId, type + obj.idbStoreSuffix, function() {
          IDB.get(userId, type + obj.idbStoreSuffix, key, function(cached) {
            callback(cached);
          },
          function(msg) {
            console.error(msg);
          });
        });
      };

      obj.add = function(userId, type, key, content) {
        cullCache(userId, type + obj.idbStoreSuffix, function() {
          content.timestamp = new Date();
          IDB.set(userId, type + obj.idbStoreSuffix, key, content, function() {
            // remove least recently used if we've hit the limit
            IDB.getGroup(userId, type + obj.idbStoreSuffix, null, null, function(records) {
              if (records && records.length > maxKeys) {
                var minTimestamp;
                for (var i = 0; i < records.length; i++) {
                  if (!minTimestamp || records[i].timestamp < minTimestamp) {
                    minTimestamp = records[i].timestamp;
                  }
                }
                IDB.deleteGroup(userId, type + obj.idbStoreSuffix, "timestamp", minTimestamp, function() {});
              }
            });
          });
        },
        function(msg) {
          console.error(msg);
        });
      };
    }
  };

  GlobalUtils.removeControlCharacters = function(str) {
    // from http://en.wikipedia.org/wiki/C0_and_C1_control_codes
    // except for \n (\u000A and \u0009) and \t for code snippets in <pre> tags
    // also the specials block
    if (!str) {
      return str;
    }
    return str.replace(/[\u0000-\u0008\u000B-\u001F\u0080-\u009F\uFFF0-\uFFFF]/g, "");
  };

  GlobalUtils.createTitleAndLinkPortionOfUrlClipContent = function(title, url) {
    var titleAttr = (title) ? GlobalUtils.escapeXML(title) : Browser.i18n.getMessage("quickNote_untitledNote");

    var content = document.createElement("div");
    content.style.whiteSpace = "nowrap";
    var t = document.createElement("div");
    t.textContent = titleAttr;
    t.style.fontFamily = "Helvetica, Arial, sans-serif";
    t.style.fontSize = "14px";
    t.style.fontWeight = "bold";
    t.style.color = "#0C0C0C";
    t.style.overflowX = "hidden";
    t.style.textOverflow = "ellipsis";
    t.style.paddingBottom = "9px";
    content.appendChild(t);
    var d = document.createElement("div");
    d.style.borderTop = "1px solid #D8D8D8";
    d.style.height = "0px";
    d.style.width = "100%";
    content.appendChild(d);
    var te = document.createElement("div");
    te.style.display = "inline-block";
    te.style.verticalAlign = "top";
    te.style.margin = "15px 0px 0px 0px";
    te.style.width = "364px";
    content.appendChild(te);
    var l = document.createElement("div");
    l.style.fontFamily = "Helvetica, Arial, sans-serif";
    l.style.fontSize = "12px";
    l.style.color = "#0C0C0C";
    l.style.display = "block";
    te.appendChild(l);
    var a = document.createElement("a");
    a.href = url;
    a.textContent = url;
    a.style.display = "inline-block";
    a.style.textDecoration = "none";
    a.style.whiteSpace = "nowrap";
    a.style.overflow = "hidden";
    a.style.textOverflow = "ellipsis";
    a.style.color = "#0C0C0C";
    a.style.width = "345px";
    l.appendChild(a);
    return { content: content, textPortion: te, link: l, url: a };
  };

  GlobalUtils.parseOperatingSystem = function(str) {
    try {
      if (/Windows/i.test(str)) {
        var version = /Windows NT (.+)/.exec(str)[1];
        switch (version) {
          case "3.1":
            return "Windows NT 3.1";
          case "3.5":
            return "Windows NT 3.5";
          case "3.51":
            return "Windows NT 3.51";
          case "4.0":
            return "Windows NT 4.0";
          case "5.0":
            return "Windows 2000";
          case "5.01":
            return "Windows 2000 SP1";
          case "5.1":
            return "Windows XP";
          case "5.2":
            return "Windows XP x64";
          case "6.0":
            return "Windows Vista";
          case "6.1":
            return "Windows 7";
          case "6.2":
            return "Windows 8";
          case "6.3":
            return "Windows 8.1";
          default:
            return "Windows";
        }
      } else if (/Mac OS X/i.test(str)) {
        return "Mac " + /OS X (.+)/.exec(str)[1].replace(/_/g, ".");
      } else if (/CrOS/i.test(str)) {
        return "Chrome OS " + /CrOS (.+)/.exec(str)[1];
      } else if (/Linux/i.test(str)) {
        return str;
      } else if (/FreeBSD/i.test(str)) {
        return str;
      }
      return null;
    } catch(err) {
      return null;
    }
  };

  GlobalUtils.generateSystemInfo = function() {
    var browser;
    if (SAFARI) {
      browser = "Safari " + /Version\/(.+?)(\s|$)/.exec(navigator.userAgent)[1];
    } else if (OPERA) {
      browser = "Opera " + /OPR\/(.+?)(\s|$)/.exec(navigator.userAgent)[1] + "/" + /Chrome\/(.+?)(\s|$)/.exec(navigator.userAgent)[1];
    } else if (YANDEX) {
      browser = "Yandex " + /YaBrowser\/(.+?)(\s|$)/.exec(navigator.userAgent)[1] + "/" + /Chrome\/(.+?)(\s|$)/.exec(navigator.userAgent)[1];
    } else { // chrome
      browser = "Chrome " + /Chrome\/(.+?)(\s|$)/.exec(navigator.userAgent)[1];
    }

    var os;
    var parReg = new RegExp("\\((.+?)\\)", "g");
    var parMatch = parReg.exec(navigator.userAgent);
    outerloop:
    while (parMatch) {
      var pieces = parMatch[1].split(/;\s?/);
      for (var i = 0; i < pieces.length; i++) {
        var os = GlobalUtils.parseOperatingSystem(pieces[i]);
        if (os) {
          break outerloop;
        }
      }
      parMatch = parReg.exec(navigator.userAgent);
    }
    if (!os) {
      os = "Unknown OS";
    }
    if (!browser) {
      browser = "Unknown browser";
    }
    return { browser: browser, os: os };
  };

  Object.preventExtensions(GlobalUtils);
})();
