/**
 * Copyright 2014 Evernote Corporation. All rights reserved.
 */

function PageInfo() {
  "use strict";

  var findArticleCallback;
  var newPageCallback;

  var clearlyDetect;
  var clearlyNext;

  // This is a map of hostnames (for hostnames that begin with 'www.', the 'www.' will be stripped off first, so don't
  // include it in your lookup string) to CSS selectors. When we try and locate an article in a page, we'll see if we
  // can find the doamin for the page in this list, and if so, we'll try and find an element that matches the given
  // selector. If no element is returned, we'll fall back to the heuristic approach.
  var specialCases = {
    "penny-arcade.com": ["div.contentArea > div.comic > img"],
    "aspicyperspective.com": ["div.entry-content"],
    "thewirecutter.com": ["div#content"],
    "katespade.com": ["div#pdpMain"],
    "threadless.com": ["section.product_section"],
    "yelp.com": ["div#bizBox"],
    "flickr.com": ["div#photo"],
    "instagr.am": ["div.stage > div.stage-inner"],
    "stackoverflow.com": ["div#mainbar"],
    "makeprojects.com": ["div#guideMain"],
    "cookpad.com": ["div#main #recipe"],
    "imgur.com": ["div.image"],
    "smittenkitchen.com": ["div.entry"],
    "allrecipes.com": ["div#content-wrapper"],
    "qwantz.com": ["img.comic"],
    "questionablecontent.net": ["img#strip"],
    "cad-comic.com": ["div#content"],
    "twitter.com": [".permalink", "div.content-main"],
    "blog.evernote.com": [".post"]
  };

  // this list of domains requires special handling in which we insert a div to
  // capture and cut out the section we want when the smallest box that holds
  // the article captures too much information. Each entry has the smallest box
  // (container) and a list of stuff in the container that we want to capture.
  var specialCasesInsertBox = {
    "blog.evernote.com": {
      container: "#page-wrap > section > article",
      include: "h2, .p-meta, .post-meta, .thumbnail, .art-content"
    },
    "kirei.biglobe.ne.jp": {
      container: "div#main",
      include: ".recipeTitle, .recipeMain, #howTo"
    },
    "nomnompaleo.com": {
      container: "section article.text",
      include: "header, section"
    },
    "foodnetwork.com": {
      container: "#fn-w",
      include: ".rcp-head, .tabnav, #recipe-lead, .w-inner, .body-text"
    }
  };

  var searchEngines = {
    Baidu: {
      regex: /^https?:\/\/([^.\/]+\.)?baidu\.(com|cn)\/s/i,
      content: "#container",
      searchBox: "input[name=wd]",
      allowedElements: "h3.t a[data-click], .f font[size='-1']",
      bannedSubelements: "span.g, .m, .c",
      titleTrim: function(title) {
        var regex = /(.+)_\u767E\u5EA6\u641C\u7D22/;
        if (regex.test(title)) {
          return regex.exec(title)[1];
        } else {
          return title;
        }
      }
    },
    Bing: {
      regex: /^https?:\/\/([^.\/]+\.)?bing\.com\/search/i,
      content: "#results_container, #b_results",
      searchBox: "input[name=q]",
      allowedElements: "div.sb_tlst a, div.sa_mc p, .b_algo h2 a, .b_algo .b_caption p"
    },
    Daum: {
      regex: /^https?:\/\/search\.daum\.net\/search/i,
      content: "#mArticle .inner_article",
      searchBox: "input[name=q]",
      allowedElements: ".coll_cont ul .wrap_tit a, .coll_cont ul .f_eb.desc",
      titleTrim: function(title) {
        var regex = /(.+)\s\u2013/;
        if (regex.test(title)) {
          return regex.exec(title)[1];
        } else {
          return title;
        }
      }
    },
    Google: {
      regex: GlobalUtils.buildGoogleRegEx(),
      content: "#rso",
      searchBox: "input[name=q][type=hidden]",
      allowedElements: "a.l, span.st", // css selector for the titles and snippets of search results
      bannedSubelements: "span.f" // css selector for nodes within allowed elements that we don't want
    },
    Naver: {
      regex: /^https?:\/\/search\.naver\.com\/search\.naver/i,
      content: "#content",
      searchBox: "input[name=query]",
      allowedElements: ".type01 dt a, " +
        ".type01 dd:not(.txt_inline):not(.txt_block):not(.review):not([style*='display:none'])",
      titleTrim: function(title) {
        var regex = /(.+)\s\u003A/;
        if (regex.test(title)) {
          return regex.exec(title)[1];
        } else {
          return title;
        }
      }
    },
    Yahoo: {
      regex: /^https?:\/\/([^.\/]+\.)*yahoo\.com\/s(earch|\?)/i,
      content: "#main",
      searchBox: "input[name=p]",
      allowedElements: ".yschttl.spt[id], .abstr"
    },
    YahooCN: {
      regex: /^https?:\/\/([^.\/]+\.)*yahoo\.cn\/s(earch|\?)/i,
      content: ".content",
      searchBox: "input[name=q]",
      allowedElements: "h3.title a, .desc",
      titleTrim: function(title) {
        var regex = /(.+)_\u7F51\u9875\u641C\u7D22/;
        if (regex.test(title)) {
          return regex.exec(title)[1];
        } else {
          return title;
        }
      }
    },
    YahooJP: {
      regex: /^https?:\/\/([^.\/]+\.)*yahoo\.co\.jp\/s(earch|\?)/i,
      content: "#WS2m ul",
      searchBox: "input[name=p]",
      allowedElements: "#WS2m .hd h3 a, #WS2m .bd p",
      bannedSubelements: "#WS2m .bd p.dlink",
      titleTrim: function(title) {
        var regex = /\u300C(.+)\u300D/;
        if (regex.test(title)) {
          return regex.exec(title)[1];
        } else {
          return title;
        }
      }
    },
    Yandex: {
      regex: /^https?:\/\/([^.\/]+\.)?yandex\.com\/yandsearch/,
      content: ".b-serp2-list.b-serp.i-bem.b-serp_js_inited.b-serp2-list_js_inited",
      searchBox: "input[name=text]",
      allowedElements: ".b-serp2-item__title-link, .b-serp2-item__text",
      bannedSubelements: ".b-serp2-item__from",
      titleTrim: function(title) {
        var regex = /(.+)\s+\u2014/;
        if (regex.test(title)) {
          return regex.exec(title)[1];
        } else {
          return title;
        }
      }
    },
    YandexRU: {
      regex: /^https?:\/\/([^.\/]+\.)?yandex\.ru\/yandsearch/,
      content: ".b-serp-list",
      searchBox: "input[name=text]",
      allowedElements: ".b-serp-item__wrapper .b-serp-item__title-link, .b-serp-item__text",
      bannedSubelements: ".b-serp-item__from",
      titleTrim: function(title) {
        var regex = /(.+)\s+\u2014/;
        if (regex.test(title)) {
          return regex.exec(title)[1];
        } else {
          return title;
        }
      }
    }
  };

  // These are the items we're trying to collect. This first block is trivial.
  var containsImages = Boolean(document.getElementsByTagName("img").length > 0);
  var documentWidth = document.width;
  var documentHeight = document.height;
  var url = document.location.href;
  var documentLength = document.body.textContent.length;

  // These take slightly more work and are initialized only when requested.
  var article = null;
  var articles = [];
  var cleanArticles = [];
  var selection = false; // This is easy to get, but is always "false" at load time until the user selects something.
  var selectionIsInFrame = false;
  var documentIsFrameset = false;
  var selectionFrameElement = null;
  var recommendationText = null;
  var favIconUrl = null;

  var clearlyInjected = false;

  // Internal state variables to keep us duplicating work.
  var hasCheckedArticle = false;

  // Hack to make Safari selection clipping work. Keeps track of the last text that was selected so that we'll still
  // have it even after the popup has appeared and erased it.
  var lastSelectionRanges = [];

  // determine if this page is a search engine page
  var searchEngine = null;
  for (var engine in searchEngines) {
    if (searchEngines[engine].regex.test(document.location.href)) {
      searchEngine = engine;
      break;
    }
  }

  // Experimental recognition of 'image' pages (like photo sites and comics).
  function findImage() {
    var imgs = document.getElementsByTagName("img");
    var biggest = null;
    var biggestArea = 0;
    for (var i = 0; i < imgs.length; i++) {
      var style = window.getComputedStyle(imgs[i]);
      var width = parseInt(style.width);
      var height = parseInt(style.height);
      var area = width * height;
      if (area > biggestArea) {
        biggest = imgs[i];
        biggestArea = area;
      }
    }
    return biggest;
  }

  function getAncestors(node) {
    var an = [];
    while (node) {
      an.unshift(node);
      node = node.parentNode;
    }
    return an;
  }

  function getDeepestCommonNode(nodeList1, nodeList2) {
    var current = null;
    for (var i = 0; i < nodeList1.length; i++) {
      if (nodeList1[i] === nodeList2[i]) {
        current = nodeList1[i];
      }
      else {
        break;
      }
    }
    return current;
  }

  function getCommonAncestor(nodeList) {
    if (!nodeList.length) return null;

    if (nodeList.length == 1) return nodeList[0];
    var lastList = getAncestors(nodeList[0]);

    var node = null;
    for (var i = 1; i < nodeList.length; i++) {
      var list = getAncestors(nodeList[i]);
      node = getDeepestCommonNode(lastList, list);
      lastList = getAncestors(node);
    }
    return node;
  }

  function clearlyCallback(data, callback) {
    // See if we should special-case this.
    var host = document.location.host;
    if (specialCasesInsertBox[host]) {
      var container = document.querySelector(specialCasesInsertBox[host].container);
      if (container) {
        var include = container.querySelectorAll(specialCasesInsertBox[host].include);
        if (include && include.length > 0) {
          var insertBefore = include[include.length - 1].nextElementSibling;
          article = document.createElement("div");
          for (var i = 0; i < include.length; i++) {
            article.appendChild(include[i]);
          }
          insertBefore.parentNode.insertBefore(article, insertBefore);
        }
      }
    }

    var useFoundImage = JSON.parse(JSClipInterface.getImageSites());
    if (!article && specialCases[host])
    {
      for (var i = 0; i < specialCases[host].length; i++) {
        var candidate = document.querySelector(specialCases[host][i]);
        if (candidate) {
          article = candidate;
          break;
        }
      }
    }

    // Or see if it's a special case image page.
    else if (useFoundImage.indexOf(host) != -1) {
      article = findImage();
    }

    // If we're scanning a search engine's page, don't grab the whole page since they contain useless words that are
    // not related to the search query.
    if (!article) {
      if (searchEngine) {
        article = document.querySelector(searchEngines[searchEngine].content);
        // no search results -> don't want it to capture the entire page, shouldn't capture anything
        if (!article || article.textContent.trim() == "") {
          article = document.createElement("div");
          document.body.insertBefore(article, document.body.firstChild);
        }
      }
    }

    // If it's not a special case, see if it's a single image.
    if (!article) {
      if (/\.(jpe?g|gif|png)$/.test(document.location.href)) {
        var candidate = document.getElementsByTagName("img")[0];
        if (candidate && candidate.parentNode === document.body) {
          article = candidate;
        }
      }
    }

    // If we still didn't find an article, let's see if maybe it's in a frame. Cleary fails on frames so we try this
    // check before we use our clearly info.
    if (!article) {
      if (document.body.nodeName.toLowerCase() == "frameset") {
        documentIsFrameset = true;
        var frame = findBiggestFrame();
        if (frame && frame.contentDocument && frame.contentDocument.documentElement) {
          selectionFrameElement = frame;
          article = frame.contentDocument.documentElement;
        }
      }
    }

    // If we didn't use any of our special case handling, we'll use whatever clearly found.
    if (!article) {
      if (data && data._elements && data._elements.length) {
        article = data._elements[0];
        if (data._elements.length > 1) {

          // This will include *all* clearly elements (and whatever else in in between them).
          article = getCommonAncestor(data._elements);

          // This includes *just the last (and therefore most important)* element from the clearly detection.
          // article = data._elements[data._elements.length - 1];
        }

        if (article.nodeType === Node.TEXT_NODE) {
          article = article.parentNode;
        }
      }
    }

    // If clearly found nothing (because it failed), then use the body of the document.
    if (!article) {
      article = document.body;
    }

    hasCheckedArticle = true;
    callback();
  }

  function afterInject() {
    // If we'd previously computed an article element, but it's lost its parent or become invisible, then we'll try
    // and re-compute the article. This can happen if, for example the page dynamically udaptes itself (like showing 
    // the latest news article in a box that updates periodically). This doesn't guarantee that we clip something 
    // sane if this happens, (if the page re-writes itself while a clip is taking place, the results are 
    // indeterminate), but it will make such things less likely.
    if (article &&
        (!article.parentNode || !article.getBoundingClientRect || article.getBoundingClientRect().width == 0)) {
      article = null;
      articles = [];
      cleanArticles = [];
      hasCheckedArticle = false;
    }

    if (!hasCheckedArticle) {
      if (typeof initClearlyComponent__detect == "function" && typeof initClearlyComponent__next == "function") {
        clearlyDetect = {
          callbacks: {
            finished: function(data) {
              if (data && data._html) {
                cleanArticles.push(data._html);
              }
              clearlyNext.start();
              clearlyCallback(data, findArticleCallback);
            },
          },
          window: window,
          jQuery: window.jQueryForClearlyComponent
        };
        clearlyDetect = initClearlyComponent__detect(clearlyDetect);
        if (clearlyDetect) {
          clearlyNext = {
            callbacks: {
              newPageFound: function(page) {
                cleanArticles.push(page._html);
                articles.push(page._elements[0]);
                if (newPageCallback) {
                  newPageCallback(page._html, page._url, cleanArticles.length);
                }
              }
            },
            settings: {
              onCreateNextPagesContainerUseThisId: "evernoteClearlyNextContainer",
            },
            detectComponentInstance: clearlyDetect
          };
          clearlyNext = initClearlyComponent__next(clearlyNext);
          if (clearlyNext) {
            clearlyNext.createNextPagesContainer();
            clearlyDetect.start();
          } else {
            console.warn("failed to initiate clearly next component");
            clearlyCallback(null, findArticleCallback);
          }
        } else {
          console.warn("failed to initiate clearly detect component");
          clearlyCallback(null, findArticleCallback);
        }
      } else {
        console.warn("Couldn't find clearly!");
        clearlyCallback(null, findArticleCallback);
      }
    }
    // If the page is big enough, clearly is excruciatingly slow. We'll jsut get the whole page.
    // @TODO: Maybe clearly can get faster.
    else if (document.body.innerHTML.length > (1024 * 1024)) {
      console.warn("Page over 1mb, skipping article detection.");
      clearlyCallback(null, findArticleCallback);
    }
    else {
      findArticleCallback();
    }
  }

  // This will try and determine the 'default' page article. It will only run once per page, but it's specifically
  // called only on demand as it can be expensive.
  function findArticle(callback) {
    findArticleCallback = callback;
    afterInject();
  }

  function findBiggestFrame() {
    var frames = document.getElementsByTagName("frame");
    var candidate = null;
    var candidateSize = 0;
    for (var i = 0; i < frames.length; i++) {
      if (frames[i].width && frames[i].height) {
        var area = frames[i].width * frames[i].height;
        if (area > candidateSize) {
          candidate = frames[i];
          candidateSize = area;
        }
      }
    }
    return candidate;
  }

  function getDefaultArticle(callback, _newPageCallback) {
    newPageCallback = _newPageCallback;
    findArticle(function(){
      callback(article);
      if (newPageCallback) {
        newPageCallback(null, null, cleanArticles.length);
      }
    });
    // Article already exists, so we'll return it.
    if (article) return article;
  }

  function getCleanArticle(callback, _newPageCallback) {
    newPageCallback = _newPageCallback;
    findArticle(function() {
      callback(cleanArticles);
    });
  }

  function getAdditionalPages(callback) {
    return articles;
  }

  Browser.addMessageHandlers({
    getInfo: getInfoRequestHandler,
    setFavIconUrl: setFavIconUrl
  });

  function rememberTextSelection() {
    var s = getSelection();
    lastSelectionRanges = [];
    if (s) {
      for (var r = 0; r < s.rangeCount; r++) {
        var range = s.getRangeAt(r);
        lastSelectionRanges[r] = range;
      }
    }
  }

  if (typeof SAFARI != "undefined" && SAFARI) {
    window.addEventListener("mouseup", rememberTextSelection);
    Browser.addKeyboardHandlers({
      "65 + 91": rememberTextSelection
    });
  }

  // Looks for selections in the current document and descendent (i)frames.
  // Returns the *first* non-empty selection.
  function getSelection() {

    // First we check our main window and return a selection if that has one. 
    var selection = window.getSelection();
    if (selection && selection.rangeCount && !selection.isCollapsed) {
      return selection;
    }

    // Then we'll try our frames and iframes.
    var docs = [];
    var iframes = document.getElementsByTagName("iframe");
    for (var i = 0; i < iframes.length; i++) {
      docs.push(iframes[i]);
    }
    var frames = document.getElementsByTagName("frame");
    for (var i = 0; i < frames.length; i++) {
      docs.push(frames[i]);
    }

    var urlBase = document.location.href.replace(/^(https?:\/\/.*?)\/.*/i, "$1").toLowerCase();
    for (var i = 0; i < docs.length; i++) {

      // If frames/iframes fail a same origin policy check, then they'll through annoying errors, and we wont be able
      // to access them anyway, so we attempt to skip anything that wont match.
      if (docs[i].src && docs[i].src.toLowerCase().substr(0, urlBase.length) !== urlBase) {
        continue;
      }

      try {
        var doc = docs[i].contentDocument;

        if (doc) {
          var frameSelection = doc.getSelection();
          if (frameSelection && frameSelection.rangeCount && !frameSelection.isCollapsed) {
            selectionIsInFrame = true;
            selectionFrameElement = docs[i];
            return frameSelection;
          }
        }
        else {
          console.warn("iframe contained no Document object.");
        }
      } catch(err) {
        // catch cross-origin error in case the check before doesn't stop it
      }
    }

    // Didn't find anything.
    return null;
  }

  function reselectText() {
    var sel = window.getSelection();
    sel.removeAllRanges();
    for (var r = 0; r < lastSelectionRanges.length; r++) {
      sel.addRange(lastSelectionRanges[r]);
    }
  }

  function getText(node, soFar, maxLen, keepText, topNode, removePunc) {
    if (node.nodeType == Node.TEXT_NODE) {
      if (keepText || !searchEngine) {
        var text = cleanSearchEngineText(node, topNode);
        // remove punctuation and whitespace
        var trimmed;
        if (removePunc) {
          trimmed = GlobalUtils.removePunctuation(text.trim()).replace(/\s+/g, " ");
        } else {
          trimmed = text.trim().replace(/\s+/g, " ");
        }
        if (trimmed === " " || trimmed === "") return soFar;
        return soFar + " " + trimmed;
      }
      return soFar;
    }

    var banned = [
      "script",
      "noscript",
      "style"
    ];

    if (node.nodeType == Node.ELEMENT_NODE) {
      if (banned.indexOf(node.nodeName.toLowerCase()) == -1) {
        // don't get the Clipper interface
        if (/^evernote.+Tools$/.test(node.id)) {
          return soFar;
        }
        for (var i = 0; i < node.childNodes.length; i++) {
          if (searchEngine) {
            // mark text within our allowed elements. keepText and topNode cascade down the DOM node subtree of topNode
            // the allowed text nodes will know to 
            if (matchesBannedSelectorInSearchEngine(node.childNodes[i])) {
              continue;
            }
            if (matchesAllowedSelectorInSearchEngine(node) || keepText) {
              if (topNode) {
                soFar = getText(node.childNodes[i], soFar, maxLen, true, topNode, removePunc);
              }
              else {
                soFar = getText(node.childNodes[i], soFar, maxLen, true, node, removePunc);
              }
            }
            else {
              soFar = getText(node.childNodes[i], soFar, maxLen, false, null, removePunc);
            }
          }
          else {
            soFar = getText(node.childNodes[i], soFar, maxLen, false, null, removePunc);
          }
          if (soFar.length > maxLen) {
            return soFar;
          }
        }
      }
    }
    return soFar;
  }

  function matchesAllowedSelectorInSearchEngine(node) {
    var selector = searchEngines[searchEngine].allowedElements;
    if (node.webkitMatchesSelector && node.webkitMatchesSelector(selector)) {
      return true;
    }
    return false;
  }

  function matchesBannedSelectorInSearchEngine(node) {
    var selector = searchEngines[searchEngine].bannedSubelements;
    if (selector) {
      if (node.webkitMatchesSelector && node.webkitMatchesSelector(selector)) {
        return true;
      }
    }
    return false;
  }

  /*
    Removes generic words from search engine result titles.
  */
  function cleanSearchEngineText(node, topNode) {
    var text = node.textContent;
    if (searchEngine) {
      var url;
      if (searchEngine == "Baidu") {
        url = topNode.parentNode.parentNode.querySelector(".g");
        if (url) {
          url = url.textContent;
        }
      }
      else if (searchEngine == "Yandex" || searchEngine == "YandexRU") {
        url = topNode.parentNode.querySelector(".b-serp-item__title-link, .serp-item__title-link");
        if (url) {
          url = url.href;
        }
      }
      else {
        url = topNode.href;
      }

      if (url) {
        var regexMap = {
          wikipedia: /(.*)[-|\u2013|\u2014]/,
          youtube: /(.*)[-|\u2013|\u2014]/,
          facebook: /(.*)\|/,
          wiktionary: /(.*)-\sWiktionary/,
          stumbleupon: /(.*)\|\sStumbleUpon\.com/
        };
        for (var site in regexMap) {
          if (new RegExp(site).test(url)) {
            var regex = regexMap[site];
            var match = regex.exec(text);
            if (match) {
              topNode.setAttribute("sawdivider", true);
              return match[1];
            }
            else if (topNode.getAttribute("sawdivider")) {
              return "";
            }
          }
        }
      }

      // remove Facebook "333 likes 666 talking about this" from search engine results
      if (/(\d+\slikes)|(\d+\stalking\sabout\sthis)/.test(text)) {
        return text.replace(/(\d+\slikes)|(\d+\stalking\sabout\sthis)/g, " ");
      }
      // remove wiktionary text
      else if (/Definition from Wiktionary, the free dictionary. Jump to: navigation, search/.test(text)) {
        return text.replace(/Definition from Wiktionary, the free dictionary. Jump to: navigation, search/g, " ");
      }
    }
    return text;
  }

  function getTitle() {
    var title = document.title;
    if (searchEngine) {
      if (searchEngines[searchEngine].titleTrim) {
        // this engine has a different method of generating result page titles than Google/Yahoo/Bing/Yandex
        title = searchEngines[searchEngine].titleTrim(title);
      }
      else {
        var pieces = title.split(" - ");
        title = title.replace(" - " + pieces[pieces.length - 1], "");
      }
    }
    return title.trim();
  }

  function getSearchQuery() {
    if (searchEngine) {
      if (document.querySelector(searchEngines[searchEngine].searchBox)) {
        var query = document.querySelector(searchEngines[searchEngine].searchBox).value.trim();
        if (query.length > 0) {
          return query;
        }
      }
    }
    return null;
  }

  function getRecommendationText(bookmark, callback) {
    var text = "";
    var MAX_LEN = 5000;
    var selection = getSelection();
    if (bookmark) {
      getDefaultArticle(function(art) {
        callback(getText(art, "", 510, false, null, false));
      });
    } else {
      if (selection) {
        var df = selection.getRangeAt(0).cloneContents();
        var div = document.createElement("div");
        div.appendChild(df);
        text = getText(div, "", MAX_LEN, false, null, true);
      }

      else if (article) {
        text = getText(article, "", MAX_LEN, false, null, true);
      }
      else {
        text = getText(document.body, "", MAX_LEN, false, null, true);
      }
      text = getTitle() + " " + text;
      return text;
    }
  }

  // Note: you must call getSelection() first to populate this field!
  function getSelectionFrame() {
    return selectionFrameElement;
  }

  function _getInfoRequestHandler(data, request, sender, sendResponse) {
    var isSelected = getSelection();

    // checkClearly();

    var response = {
      containsImages: containsImages,
      documentWidth: documentWidth,
      documentHeight: documentHeight,
      url: url,
      selection: (isSelected !== null) || (lastSelectionRanges.length > 0),
      selectionIsInFrame: selectionIsInFrame,
      documentLength: document.body.textContent.length,
      article: (article != null),
      recommendationText: getRecommendationText(false),
      query: getSearchQuery(),
      searchEngine: searchEngine,
      favIconUrl: favIconUrl,
      documentIsFrameset: documentIsFrameset,
      pdf: getPdfUrl(),
      gmailThread: isGmailThread(),
      gmail: isGmail()
    };

    sendResponse(response);
  }

  function respondWithInfo(info) {
    Browser.sendToExtension({name: "simSearch_receivePageInfo", info: info});
  }

  function getInfoRequestHandler(request, sender, sendResponse) {
    // hack for safari to clip a selection from the popover. a bug in safari erases the text selection everytime the
    // popover comes up. so reselect it after the popover appears.
    if (typeof SAFARI != "undefined" && SAFARI && !request.sendToTab && lastSelectionRanges.length > 0
        && getSelection() == null) {
      reselectText();
    }
    findArticle(function(data){_getInfoRequestHandler(data, request, sender, respondWithInfo)});
  }

  function getPdfUrl() {
    if (document.querySelector("embed[type='application/pdf']")) {
      return document.querySelector("embed[type='application/pdf']").src;
    } else if (/^https?:\/\/docs\.google\.com\/viewer\?url=.+/.test(document.location.href)) {
      for (var s = 0; s < document.scripts.length; s++) {
        if (/gviewApp\.setFileData/.test(document.scripts[s].innerText)) {
          if (/mimeType.+application\/pdf/.test(document.scripts[s].innerText)) {
            var m = /^https?:\/\/docs\.google\.com\/viewer\?url=(.+?)(?:$|&)/.exec(document.location.href);
            return decodeURIComponent(m[1]);
          }
          break;
        }
      }
    } else if (getRealUrlFromOperaPdfViewerUrl()) {
      return getRealUrlFromOperaPdfViewerUrl();
    }
    return null;
  }

  function isGmail() {
    if (/^https:\/\/mail\.google\.com\/mail\//.test(document.location.href)) {
      return true;
    }
    return false;
  }

  function isGmailThread() {
    if (isGmail()) {
      var threadButtons = document.querySelectorAll("span > div > span > [src='images/cleardot.gif']");
      if (threadButtons.length > 0) {
        return true;
      }
    }
    return false;
  }

  function setFavIconUrl(request, sender, sendResponse) {
    favIconUrl = request.favIconUrl;
  }

  function getBiggestImage(callback) {
    getDefaultArticle(function(art) {
      var imgs;
      if (art) {
        imgs = art.querySelectorAll("img");
      } else {
        imgs = document.querySelectorAll("img");
      }

      var maxHeight = 0;
      var maxWidth = 0;
      var maxImage;
      for (var i = 0; i < imgs.length; i++) {
        var w = imgs.item(i).width;
        var h = imgs.item(i).height;
        if (w * h > maxWidth * maxHeight) {
          maxHeight = h;
          maxWidth = w;
          maxImage = imgs.item(i).src;
        }
      }
      callback({ src: maxImage, width: maxWidth, height: maxHeight });
    });
  }

  function getCanonicalUrl() {
    if (!isGmail()) {
      var link = document.querySelector("link[rel='canonical']");
      if (link) {
        return link.href;
      }
    }
    return null;
  }

  function getRealUrlFromOperaPdfViewerUrl() {
    var match = /^chrome-extension:\/\/(?:encfpfilknmenlmjemepncnlbbjlabkc|oemmndcbldboiebfnladdacbdfmadadm)\/(.+)/.exec(document.location.href);
    if (match) {
      return match[1];
    }
    return null;
  }

  // Public API:
  this.getBiggestImage = getBiggestImage;
  this.getDefaultArticle = getDefaultArticle;
  this.getCanonicalUrl = getCanonicalUrl;
  this.getCleanArticle = getCleanArticle;
  this.getAdditionalPages = getAdditionalPages;
  this.getSelection = getSelection;
  this.getSelectionFrame = getSelectionFrame;
  this.getPdfUrl = getPdfUrl;
  this.getRecommendationText = getRecommendationText;
  this.getSearchQuery = getSearchQuery;
  this.getTitle = getTitle;
  this.isGmail = isGmail;
  this.isGmailThread = isGmailThread;
  this.getRealUrlFromOperaPdfViewerUrl = getRealUrlFromOperaPdfViewerUrl;
  this.__defineGetter__("documentIsFrameset", function() {
    return documentIsFrameset;
  });
  this.__defineGetter__("favIconUrl", function() {
    return favIconUrl;
  });

  return this;
}
