/* 
 * How this HTML serializer works:
 *
 * The serialization of the document structure itself in here is fairly straightforward: it's a recursive descent
 * parser that will serialize HTML into ENML-ready markup. It converts unknown elements into generic equivalents, drops
 * prohibited elements, copies attributes across (unless they're restricted) and basically does what you'd expect, with
 * the following exceptions:
 * 
 * It breaks every N nodes serialized and fires off a callback (via setTimeout) to resume in a few milliseconds.
 * This has a couple important properties. It keeps the UI from blocking completely while the serializer runs (but
 * still makes it sluggish), and more importantly, it keeps browsers from deciding that the script on the page is
 * spinning and prompting the user to kill it.
 * 
 *
 * CSS: The interesting portion of how this whole things works is how it goes about inlining stylesheets. The core of
 * this is built on an API called getMatchedCSSRules (webkit-only, for other browsers, there is another implementation
 * in here built around matchesSelector, which is both much slower and probably not kept in sync with the primary
 * version). 
 * 
 * Calling getMatchedCSSRules will give you a list of all the CSS rules that apply to your element, but with some
 * caveats: noticeably, you wont get back rules from stylesheets that you wouldn't normally be able to load yourself
 * due to the same-origin policy. These are just silently omitted. To get around this, we start our serialization
 * process by iterating across all the stylesheets attached to the document, and if they are link elements with 'href'
 * properties, we'll pass a message to the extension page asking it to request that URL on our behalf (because the
 * background page in an extension is not subject to the same origin policy). We don't need to bother for styles in
 * "style" tags because they can't have same-origin policy issues. We also run through each rule in each stylesheet and
 * if it's an @import rule, we'll send off a request for that stylesheet as well. We keep a list of stylesheets
 * requested and wont request duplicates, and will give up at a max of 100 total stylesheets (fun fact: @import rules
 * can be circular).
 * 
 * The background page will return us a message with the result from each requested stylesheet. We create a "style"
 * tag, and then insert the text from the request in there, and append the new style element to the page. We then fire
 * off a polling function that waits for a new item to show up in document.styleSheets that references the style tag we
 * just added. When that shows up, we'll run the newly attached styleSheet object through our @import check again to
 * see if we need to grab any more stylesheets from there.
 *
 * If the polling function gets to the point that there are no more outstanding requests, it stops its own polling and
 * fires off the recursive serializer.
 *
 * At each node in the tree, we grab the list of matching CSS rules, which will now include all the ones from the style
 * tags that we added from originally third-party stylesheets, and grab the CSS *text* from each one, rather than the
 * parsed CSS representation. This will let us keep properties like "border" without expanding it out into
 * "border-(top|bottom|left|right)", and it will also keep "invalid" styles that the webkit parser would otherwise drop
 * (like properties beginning with "-moz-"). We parse this CSS text into name/value pairs, and expand out any URLs
 * referenced into absolute paths (important note, relative paths need to be made absolute relative to the path of the
 * stylesheet, not the document).
 * 
 * As we iterate through each matching rule, we let matching properties from later rules overwrite matched properties
 * from earlier rules. This may not strictly be correct, the order in which matching rules are returned is unspecified,
 * ideally we could sort them, but the logic for doing this is actually non-trivial, and it rarely makes a difference
 * anyway.
 *
 * Once we finish checking each matched CSS rule, we have a final name/value map containing all the CSS properties that
 * we want to keep. We serialize this into a "style" attribute and add it to our element, and then move on to serialize
 * its children.
 */

// Shim 'SAFARI' into existence for scripts running in frame contexts.
if (typeof SAFARI == "undefined") {
  SAFARI = false;
}

// Shim 'Browser' into existence for scripts running in frame contexts.
if (typeof Browser == "undefined") {
  Browser = {};
  Browser.sendToExtension = function(msg) {
    window.parent.postMessage(msg, "*");
  };

  var msgMap = {};

  var msgHandler = function(msg) {
    var data = msg.data;
    if (data && data.name && msgMap[data.name]) {
      msgMap[data.name](msg.data);
    }
  };
  window.addEventListener("message", msgHandler, false);

  Browser.addMessageHandlers = function(obj) {
    for (var i in obj) {
      msgMap[i] = obj[i];
    }
  };
}

function HtmlSerializer() {
  "use strict";

  function clipArticle(serializer, complete, forceWholePage) {
    var el;

    if(forceWholePage) {
      serializer.serialize(document.body, null, true, complete);
      return;
    }

    try {
    console.log("calling getDefaultArticle in html_serializer.js");
      evernotePageInfo.getDefaultArticle(function proxy(article) {
                                   console.log("article="+article);
                                   serializer.serialize(article, null, true, complete);
                                 });
      return;
    }
    catch (e) {
      console.log("Couldn't get article from pageInfo. Trying default: "+e);
    }

    serializer.serialize(document.body, null, true, complete);
  }

  this.clipArticle = clipArticle;

  var pendingStyleSheets = [];
  var counters = {};

  var pendingStyleCount = 0;
  var styleSheetList = [];
  var element;
  var range;
  var keepStyle;
  var callbacks = [];
  var stylesToRemove = [];
  var iterationCount = 0;
  var stack = [];
  var blocked = false;
  var pseudoElementRules = [];
  var frameData;
  var _window;

  // getMatchedCSSRules seems to fail on pages with a base specified, so if we find a base tag, then we temporarily
  // remove it from the document and store it here. This only works in Chrome, in Safari getMatchedCSSRules still
  // doesn't work, even if we remove this, so there is a separate check for Safari, so that we won't use
  // getMatchedCSSRules on pages with BASE tags.
  var base = null;

  var maxStylesToAdd = 100;
  var stylesAdded = [];

  var allowedElements = [
    "A",
    "ABBR",
    "ACRONYM",
    "ADDRESS",
    "AREA",
    "B",
    "BDO",
    "BIG",
    "BLOCKQUOTE",
    "BR",
    "CAPTION",
    "CENTER",
    "CITE",
    "CODE",
    "COL",
    "COLGROUP",
    "DD",
    "DEL",
    "DFN",
    "DIV",
    "DL",
    "DT",
    "EM",
    "FONT",
    "H1",
    "H2",
    "H3",
    "H4",
    "H5",
    "H6",
    "HR",
    "I",
    "IMG",
    "INS",
    "KBD",
    "LI",
    "MAP",
    "OL",
    "P",
    "PRE",
    "Q",
    "S",
    "SAMP",
    "SMALL",
    "SPAN",
    "STRIKE",
    "STRONG",
    "SUB",
    "SUP",
    "TABLE",
    "TBODY",
    "TD",
    "TFOOT",
    "TH",
    "THEAD",
    "TR",
    "TT",
    "U",
    "UL",
    "VAR"
  ];

  var html5Elements = [
    "ARTICLE",
    "ASIDE",
    "DETAILS",
    "FOOTER",
    "FIGURE",
    "FIGCAPTION",
    "HEADER",
    "HGROUP",
    "NAV",
    "SECTION",
    "SUMMARY",
  ];

  var disallowedElements = [
    "APPLET",
    "BASE",
    "BASEFONT",
    "BGSOUND",
    "BLINK",
    "BODY",
    "BUTTON",
    "DIR",
    "EMBED",
    "FIELDSET",
    "FORM",
    "FRAME",
    "FRAMESET",
    "HEAD",
    "HTML",
    "IFRAME",
    "ILAYER",
    "INPUT",
    "ISINDEX",
    "LABEL",
    "LAYER",
    "LEGEND",
    "LINK",
    "MARQUEE",
    "MENU",
    "META",
    "NOEMBED", /* ENML doesn't support this, so we drop it to keep it from rendering, since it's unlikely to apply. */
    "NOFRAMES",
    "NOSCRIPT",
    "OBJECT",
    "OPTGROUP",
    "OPTION",
    "PARAM",
    "PLAINTEXT",
    "SCRIPT",
    "SELECT",
    "STYLE",
    "TEXTAREA",
    "TITLE", // should not appear in the body
    "XML"
  ];

  // In addition to the following, any attribute beginning with "on" is disallowed.
  var disallowedAttributes = [
    "id",
    "class",
    "accesskey",
    "data",
    "dynsrc",
    "tabindex",
    "style" // We strip style attributes because we build our own.
  ];

  // from http://xml.evernote.com/pub/enml2.dtd
  var coreAttrs = [
    "style",
    "title"
  ];

  var i18nAttrs = [
    "lang",
    "xml:lang",
    "dir"
  ];

  var focusAttrs = [
    "accesskey",
    "tabindex"
  ];

  var attrs = [
    "style",
    "title",
    "lang",
    "xml:lang",
    "dir"
  ];

  var cellHAlignAttrs = [
    "align",
    "char",
    "charoff"
  ];

  // Properties we'll strip from ancestor elements to the main element being serialized. We want to keep inheritable
  // properties on these, like fonts and colors, but lose positioning.
  var strippableProperties = [
    "border",
    "border-bottom",
    "border-bottom-color",
    "border-bottom-style",
    "border-bottom-width",
    "border-collapse",
    "border-color",
    "border-left",
    "border-left-color",
    "border-left-style",
    "border-left-width",
    "border-right",
    "border-right-color",
    "border-right-style",
    "border-right-width",
    "border-spacing",
    "border-style",
    "border-top",
    "border-top-color",
    "border-top-style",
    "border-top-width",
    "border-width",
    "bottom",
    "clear",
    "display",
    "float",
    "height",
    "layout-flow",
    "layout-grid",
    "layout-grid-char",
    "layout-grid-char-spacing",
    "layout-grid-line",
    "layout-grid-mode",
    "layout-grid-type",
    "left",
    "margin",
    "margin-bottom",
    "margin-left",
    "margin-right",
    "margin-top",
    "max-height",
    "max-width",
    "min-height",
    "min-width",
    "padding",
    "padding-bottom",
    "padding-left",
    "padding-right",
    "padding-top",
    "position",
    "right",
    "size",
    "table-layout",
    "top",
    "visibility",
    "width",
    "z-index"
  ];

  function attributeAllowed(tagName, attrName) {
    attrName = attrName.toLowerCase();
    if (attrName.match(/^on/i)) return false;
    if (attrName.match(/^data-/i)) return false;
    if (disallowedAttributes.indexOf(attrName) > -1) return false;

    // strip attributes specific to tags
    // spec taken from http://xml.evernote.com/pub/enml2.dtd
    switch(tagName) {
      case "a":
        if (attrs.indexOf(attrName) > -1
            || focusAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["charset", 
            "coords", 
            "href", 
            "hreflang", 
            "name", 
            "rel", 
            "rev", 
            "shape", 
            "target", 
            "type"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "abbr":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "acronym":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "address":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "area":
        if (attrs.indexOf(attrName) > -1
            || focusAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["alt", 
            "coords", 
            "href", 
            "nohref", 
            "shape", 
            "target"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "b":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "bdo":
        if (coreAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["dir", 
            "lang"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "big":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "blockquote":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["cite"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "br":
        if (coreAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["clear"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "caption":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["align"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "center":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "cite":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "code":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "col":
        if (attrs.indexOf(attrName) > -1
            || cellHAlignAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["span", 
            "valign", 
            "width"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "colgroup":
        if (attrs.indexOf(attrName) > -1
            || cellHAlignAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["valign"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "dd":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "del":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["cite", 
            "datetime"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "dfn":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "div":
        if ([ "evernote_attachment_url", "evernote_attachment_name" ].indexOf(attrName) > -1) {
          return true;
        }
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["align"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "dl":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["compact"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "dt":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "em":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "font":
        if (coreAttrs.indexOf(attrName) > -1
            || i18nAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["color", 
            "face", 
            "size"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "h1":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["align"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "h2":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["align"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "h3":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["align"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "h4":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["align"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "h5":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["align"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "h6":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["align"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "hr":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["align", 
            "noshade", 
            "size", 
            "width"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "i":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "img":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["align", 
            "alt", 
            "border", 
            "height", 
            "hspace", 
            "ismap", 
            "longdesc", 
            "name", 
            "src", 
            "usemap", 
            "vspace", 
            "width"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "ins":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["cite", 
            "datetime"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "kbd":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "li":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["type", 
            "value"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "map":
        if (i18nAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["name", 
            "title"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "ol":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["compact", 
            "start", 
            "type"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "p":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["align"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "pre":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["width"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "q":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["cite"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "s":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "samp":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "small":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "span":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "strike":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "strong":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "sub":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "sup":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "table":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["align", 
            "bgcolor", 
            "border", 
            "cellpadding", 
            "cellspacing", 
            "summary", 
            "width"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "tbody":
        if (attrs.indexOf(attrName) > -1
            || cellHAlignAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["valign"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "td":
        if (attrs.indexOf(attrName) > -1
            || cellHAlignAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["abbr", 
            "bgcolor", 
            "colspan", 
            "height", 
            "nowrap", 
            "rowspan", 
            "valign", 
            "width"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "tfoot":
        if (attrs.indexOf(attrName) > -1
            || cellHAlignAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["valign"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "th":
        if (attrs.indexOf(attrName) > -1
            || cellHAlignAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["abbr", 
            "bgcolor", 
            "colspan", 
            "height", 
            "nowrap", 
            "rowspan", 
            "valign", 
            "width"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "thead":
        if (attrs.indexOf(attrName) > -1
            || cellHAlignAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["valign"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "tr":
        if (attrs.indexOf(attrName) > -1
            || cellHAlignAttrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["bgcolor", 
            "valign"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "tt":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "u":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "ul":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        if (["compact", 
            "type"].indexOf(attrName) > -1) {
          return true;
        }
        return false;
      case "var":
        if (attrs.indexOf(attrName) > -1) {
          return true;
        }
        return false;
    }
  }

  function nodeAllowed(nodeName) {
    nodeName = nodeName.toUpperCase();
    return (disallowedElements.indexOf(nodeName) == -1);
  }

  function transformNode(node) {
    var nodeName = node.nodeName;
    nodeName = nodeName.toUpperCase();

    if (nodeName == "INPUT" && node.type && node.type.toLowerCase() == "image") {
      return "img";
    }

    // If there's special handling for this type, put it here.
    if (nodeName == "BODY") return "div";
    if (nodeName == "HTML") return "div";
    if (nodeName == "FORM") return "div";
    if (nodeName == "LABEL") return "span";
    if (nodeName == "FIELDSET") return "div";
    if (nodeName == "LEGEND") return "span";
    if (nodeName == "IFRAME") return "div";
    if (nodeName == "EMBED") return "div";
    if (nodeName == "CANVAS") return "img";
    if (nodeName == "VIDEO") return "img";
    if (nodeName == "HIGHLIGHT") return "span";
    // These HTML5 tags should be rendered as divs instead of spans
    if (html5Elements.indexOf(nodeName) > -1) {
      return "div";
    }

    // If the node's not allowed, we want to make sure we reutn it as is, so nothing thinks it's supposed to be the
    // transformed type.
    if (!nodeAllowed(nodeName)) {
      return nodeName.toLowerCase();
    }

    // If it's not specifically allowed, either, then we'll turn it into a span, this preserves the content of special
    // node types from HTML5 and the future.
    if (allowedElements.indexOf(nodeName) == -1) {
      return "span";
    }

    // Anything else gets returned as is.
    return nodeName.toLowerCase();
  }

  Browser.addMessageHandlers({
    content_textResource: msgHandlerTextResource
  });

  // in some browsers, attempting to touch the cssRules result in a SecurityError
  function canAccessCssRules(sheet) {
    try {
      sheet.cssRules; // attempt to access it
    } catch (exception) {
      return false;
    }
    return true;
  }

  function serialize(_element, _range, _keepStyle, callback, _frameData, win) {
    try {
      _window = win;
      if (!_window) {
        _window = window;
      }
      if (callback) {
        callbacks.push(callback);
      }
      if (!blocked) {
        blocked = true;
        element = _element;
        range = _range;
        keepStyle = _keepStyle;
        frameData = _frameData;
        checkStyleSheets();
      }
      else {
        console.warn("Called serialize while blocked. Added callback but won't change base element.");
      }
    } catch(err) {
      doneRecursing(null, err);
    }
  }

  function checkStyleSheet(sheet) {
    if (!usableMedia(sheet)) {
      return;
    }
    if (stylesAdded.length >= maxStylesToAdd) {
      console.warn("Hit style cap of " + maxStylesToAdd + " styles. Stopping.");
      return;
    }

    // This is probably third party, which is why we can't read the rules (because of the same origin policy).
    if ((!canAccessCssRules(sheet) || !sheet.cssRules) && sheet.href) {
      styleSheetList.push({href: sheet.href, owner: sheet.ownerNode});
      pendingStyleCount++;
      stylesAdded.push(sheet.href);

      pendingStyleSheets.push(sheet.href);
      Browser.sendToExtension({name: "main_getTextResource", href: sheet.href});
      return;
    }

    // Prepend any @imports.
    var rules = sheet.cssRules;
    // import rule's stylesheet could've been from a different domain/path, so
    // use that domain/path instead of the current page's domain/path
    var path = null;
    if (sheet.ownerNode && sheet.ownerNode.dataset && sheet.ownerNode.dataset["evernoteOriginatingUrl"]) {
      var pathMatch = /(.+)\//.exec(sheet.ownerNode.dataset["evernoteOriginatingUrl"]);
      if (pathMatch) {
        path = pathMatch[1];
      }
    }
    for (var j = 0; j < rules.length; j++) {
      if (rules[j].type == CSSRule.IMPORT_RULE) {
        if (!rules[j].styleSheet) {
          continue;
        }
        var href = rules[j].styleSheet.href;
        if (!/^https?:\/\//.test(rules[j].href) && !/^\/\//.test(rules[j].href)) {
          if (path) {
            if (rules[j].href[0] === "/") {
              href = path + rules[j].href;
            } else {
              href = path + "/" + rules[j].href;
            }
          }
        }
        if (stylesAdded.indexOf(href) != -1) {
          continue; // Duplicate (a fun case was when these were circular).
        }

        styleSheetList.push({ href: href, owner: sheet.ownerNode });
        pendingStyleCount++;
        stylesAdded.push(href);
        pendingStyleSheets.push(href);
        Browser.sendToExtension({ name: "main_getTextResource", href: href });
      } else if (rules[j].type == CSSRule.MEDIA_RULE) {
        if (usableMedia(rules[j])) {
          styleSheetList.push(rules[j]);
        }
      }
    }

    // If we get this far, we want to keep this stylesheet, too.
    styleSheetList.push(sheet);
  }

  function checkStyleSheets() {
    pendingStyleCount = 0;
    stylesAdded = [];
    counters = {};
    base = document.getElementsByTagName("base")[0];
    for (var i = 0; i < _window.document.styleSheets.length; i++) {
      checkStyleSheet(_window.document.styleSheets[i]);
    }
    if (pendingStyleCount == 0) {
      startRecurse(element);
    }
  }

  function reconstituteUrl(base, url) {
    var reconstituted;
    url = url.trim(); // for cases like: url( http://www.com/ )
    if (url.match(/^http/i)) {
      reconstituted = url;
    }
    else if (url.match(/^\//)) {
      reconstituted = base.replace(/^(.*?:\/\/[^\/]+).*$/, "$1") + url;
    }
    else {
      reconstituted = base.replace(/^(.*\/)/, "$1") + url;
    }
    reconstituted = "url('" + reconstituted + "')";
    return reconstituted;
  }

  // We need to make our style tag match our original stylesheet.
  function preProcessStyle(styleText, originatingSheetHref) {

    var pageBase = _window.document.location.href.replace(/[^\/]+$/, "");
    var styleBase = originatingSheetHref.replace(/[^\/]+$/, "");

    if (pageBase == styleBase) {
      return styleText;
    }

    // This first block repairs URL paths.  
    // call reconstituteUrl, but prepend the original base URL to its arguments list.
    function reconstitute() {
      var args = [styleBase];
      for (var i = 0; i < arguments.length; i++) args.push(arguments[i]);
      return reconstituteUrl.apply(this, args);
    }
    if (styleText) {
      styleText = styleText.replace(/url\(["']?(.*?)["']?\)/g, reconstitute);
    }

    return styleText;
  }

  var outstandingStyleSheets = [];
  var styleInterval = 0;
  var elapsedCSSTime = 0;

  var CSS_POLL_INTERVAL = 100;
  var CSS_POLL_MAXIMUM = 10000;

  function pollForStyleSheets(){
    if (styleInterval) return;
    styleInterval = setInterval(function() {
      try {
        if (stylesAdded.length >= maxStylesToAdd || elapsedCSSTime > CSS_POLL_MAXIMUM) {
          console.log('cancelling css poll');
          cancelPollForStyleSheets();
          startRecurse(element);
          return;
        }

        OUTER: for (var i = 0; i < outstandingStyleSheets.length; i++) {
          var style = outstandingStyleSheets[i][0];
          var idx = outstandingStyleSheets[i][1];
          for (var j = 0; j < _window.document.styleSheets.length; j++) {
            var sheet = _window.document.styleSheets[j];

            if (sheet.ownerNode === style) {
              styleSheetList[idx] = sheet;
              checkStyleSheet(_window.document.styleSheets[j]);
              outstandingStyleSheets.splice(i, 1);
              pendingStyleCount--;
              break OUTER;
            }
          }
        }
        if (pendingStyleCount == 0) {
          cancelPollForStyleSheets();
          startRecurse(element);
        }
        elapsedCSSTime += CSS_POLL_INTERVAL;
      } catch(err) {
        doneRecursing(null, err);
      }
    }, CSS_POLL_INTERVAL);
  }

  function cancelPollForStyleSheets() {
    if (styleInterval) {
      clearInterval(styleInterval);
      styleInterval = 0;
    }
  }

  function msgHandlerTextResource(request, sender, sendResponse) {
    try {
      var sheetIndex = pendingStyleSheets.indexOf(request.href);
      if (sheetIndex == -1) {
        if (pendingStyleSheets.length) {
          pollForStyleSheets();
          return;
        }
      }
      else {
        pendingStyleSheets.splice(sheetIndex, 1);
      }

      for (var i = 0; i < styleSheetList.length; i++) {
        var sheet = styleSheetList[i];

        if (sheet.href === request.href) {
          var styleText = preProcessStyle(request.responseText, sheet.href);
          var style = _window.document.createElement("style");
          style.type = "text/css";
          style.textContent = styleText;
          style.dataset["evernoteOriginatingUrl"] = sheet.href;

          // Save the current list, as the new style will get inserted at some unknown point in the list, and we'll have
          // to pick it out.
          var savedStyles = [];
          for (var j = 0; j < _window.document.styleSheets.length; j++) {
            savedStyles.push(_window.document.styleSheets[j]);
          }

          if (sheet.owner) {
            sheet.owner.parentNode.insertBefore(style, sheet.owner);
          }
          else {
            _window.document.head.appendChild(style);
          }
          stylesToRemove.push(style);

          outstandingStyleSheets.push([style, i]);

          break;
        }
      }
      pollForStyleSheets();
    } catch(err) {
      doneRecursing(null, err);
    }
  }

  // "screen" and "all" are usable, and no media at all is a default of "screen".
  function usableMedia(stylesheet) {
    if (stylesheet.media && stylesheet.media.length) {
      for (var j = 0; j < stylesheet.media.length; j++) {
        var m = stylesheet.media[j].toLowerCase();
        if (m.match(/\bscreen\b/i) || m.match(/\ball\b/i)) {
          return true;
        }
      }
      return false;
    }
    return true;
  }

  function postProcessStyles() {
    pseudoElementRules = [];
    for (var i = 0; i < styleSheetList.length; i++) {
      var sheet = styleSheetList[i];
      if (!canAccessCssRules(sheet) || !sheet.cssRules) {
        continue;
      }
      for (var j = 0; j < sheet.cssRules.length; j++) {
        var rule = sheet.cssRules[j];
        if (rule.selectorText && rule.selectorText.match(/(:?:before)|(:?:after)/)) {
          pseudoElementRules.push(rule);
        }
      }
    }
  }

  function startRecurse(el) {
    // Styles don't seem to get resolved instantly all the time, so we introduce a small delay and hope that helps.
    setTimeout(function() {
      try {
        postProcessStyles();
        stack = [];
        stack.push({element: el, string: "", i: 0, after: null});
        recurse();
      } catch(err) {
        doneRecursing(null, err);
      }
    }, 300);
  }

  function escapeHTML(str) {
    return GlobalUtils.escapeXML(str);
  }

  function shouldSerializeYoutubeVideo(el) {
    // svg elements' class names are not strings...they're SVGAnimatedString
    return typeof el.id == "string" && (el.id == "movie_player")
      && _window.document.location.href.match(/v=(.*?)(&|$)/);
  }

  function serializeYoutubeVideo(el) {
    if (shouldSerializeYoutubeVideo(el)) {
      var vidId = _window.document.location.href.match(/v=(.*?)(&|$)/)[1];
      return "<a href=\"" + GlobalUtils.escapeXML(_window.document.location.href) + "\" target=\"_blank\">"
        + "<img src=\"http://img.youtube.com/vi/" + vidId + "/0.jpg\"></img></a>";
    }
    return "";
  }

  function addPicEmbeddedYoutubeVideo(el) {
    try {
      // show video's image for embedded youtube videos
      if (el.getAttribute("type") === "application/x-shockwave-flash" && el.getAttribute("flashvars")) {
        var match = /iurlsd=(.+?)(&|$)/.exec(el.getAttribute("flashvars"));
        if (!match) {
          match = /iurl=(.+?)(&|$)/.exec(el.getAttribute("flashvars"));
        }
        if (match) {
          el.style.backgroundImage = "url(" + decodeURIComponent(match[1]) + ")";
          el.style.backgroundPosition = "50%";
          el.style.backgroundRepeat = "no-repeat";
        }
        // add the link to the video
        match = /video_id=(.+?)(&|$)/.exec(el.getAttribute("flashvars"));
        if (match) {
          var ytLink = document.createElement("a");
          ytLink.href = "https://www.youtube.com/watch?v=" + match[1];
          ytLink.target = "_blank";
          ytLink.style.display = "block";
          ytLink.style.width = "100%";
          ytLink.style.height = "100%";
          el.appendChild(ytLink);
        }
      }
    } catch(err) {
      console.warn("problem with adding pic to embedded youtube video");
    }
  }

  function serializeIframe(el) {
    if (el.nodeName.toLowerCase() != "iframe") {
      return null;
    }
    var contents = "";
    if (el.dataset && el.dataset.en_id) {
      if (frameData && frameData[el.dataset.en_id]) {
        contents = frameData[el.dataset.en_id];
        var style = resolveStyle(el);
        if (el.width) style.map.width = {value: el.width + "px"};
        if (el.height) style.map.height = {value: el.height + "px"};
        // if there isn't a position attribute specified or it's static, use
        // relative to simulate iframe and contents of iframe would sit well
        // inside the "frame", e.g. divs inside the iframe that have height
        // 100% wouldn't fill 100% of the page, just this "frame"
        if (!style.map.position || style.map.position.value === "static") {
          style.map.position = { value: "relative" };
        }

        var styleStr = "";
        for (var i in style.map) {
          styleStr += i + ":" + style.map[i].value + ";";
        }

        if (styleStr) {
          styleStr = " style=\"" + escapeAttr(styleStr) + "\"";
        }
        else {
          styleStr = "";
        }

        var div = "<div" + styleStr + ">" + contents + "</div>";
        return div;
      }
    }
    return null;
  }

  // If we try to serialize a DL that contains elements other than DD or DT, the server will try to coerce this into
  // valid HTML by auto-closing our list. Instead, we simply discard invalid elements here.
  // This function returns true if this is a valid DD or DT or isn't a child of a DL. It returns false if this is an
  // invalid DL child.
  function checkValidDlChild(el) {
    if (el.parentNode && el != element) {
      var parentName = el.parentNode.nodeName;
      parentName = parentName.toLowerCase();
      if (parentName == "dl") {
        var normalized = transformNode(el);
        if (normalized != "dd" && normalized != "dt") {
          return false;
        }
      }
    }
    return true;
  }

  function recurse() {
    iterationCount++;
    if (iterationCount % 500 == 0) {
      setTimeout(function() {
        try {
          recurse();
        } catch(err) {
          doneRecursing(null, err);
        }
      }, 25);
      return;
    }

    // We're not allowed to create variables in here (they need to be on our stack), but this is just a convenience
    // mapping. (f == frame);
    var f = stack[stack.length - 1];
    if (!f) {
      // This can occur in strange cases, like we have a page that uses frames, but the frame we would choose as the
      // main content area fails the same-origin-policy check. We'll get a blank note, but at least not hang.
      doneRecursing("");
    }

    // We haven't gotten into any of our children yet.
    if (f.i == 0) {

      // This particular 'if' block isn't interruptable. Any vars declared in here must only live in here.
      if (!nodeAllowed(transformNode(f.element))) {
        stack.pop();
        recurse();
        return;
      }

      if (!checkValidDlChild(f.element)) {
        console.warn("discarding invalid DL child \"" + f.element.nodeName + "\"");
        stack.pop();
        recurse();
        return;
      }

      if (range && f.element != range.commonAncestorContainer && !range.intersectsNode(f.element)) {
        stack.pop();
        recurse();
        return;
      }

      // Drop trailing paragraphs that you get when you triple-click to select.
      if (range && f.element === range.endContainer && range.endOffset === 0) {
        stack.pop();
        recurse();
        return;
      }

      // Don't attempt to serialize the UI frames
      if (["evernoteClipperResult", "evernoteAuthTools", "evernoteFilingTools",
        "evernoteGlobalTools", "evernoteUserTools", "evernoteShareTools",
        "evernoteEmailSharing", "evernoteOptionsPage", "evernoteClearlyArticle"].indexOf(f.element.id) > -1) {
        stack.pop();
        recurse();
        return;
      }

      var ytvid = serializeYoutubeVideo(f.element);
      if (ytvid) {
        f.string += ytvid;
        stack.pop();
        if (stack.length) {
          stack[stack.length - 1].string = f.string;
          recurse();
        }
        else {
          doneRecursing(f.string);
        }
        return;
      }

      var iframe = serializeIframe(f.element);
      if (iframe) {
        f.string += iframe;
        stack.pop();
        if (stack.length) {
          stack[stack.length - 1].string = f.string;
          recurse();
        }
        else {
          doneRecursing(f.string);
        }
        return;
      }

      var style = {};
      if (keepStyle) {
        addPicEmbeddedYoutubeVideo(f.element);
        if (f.element.nodeName === "HIGHLIGHT") {
          style = { style: " style=\"x-evernote:highlighted;background-color:#F6EE96;\"" };
        } else {
          style = resolveStyle(f.element);
        }
        if (style.after) f.after = style.after;
      }

      if (style.map && style.map.display && style.map.display.value == "none") {
        // Skipping hidden element.
        stack.pop();
        recurse();
        return;
      }

      var nodeName = transformNode(f.element);
      f.string += "<" + nodeName;

      specifyImgDims(f.element);
      if (f.element.nodeName == "CANVAS") {
        // convert canvases to imgs
        try {
          f.string += " src=\"" + f.element.toDataURL() + "\"";
        } catch (err) { /* canvas was tainted, so ignore */ }
      } else if (f.element.nodeName == "VIDEO") {
        if (f.element.poster) {
          f.string += " src=\"" + GlobalUtils.escapeXML(f.element.poster) + "\"";
        }
      }
      if (f.element.attributes && f.element.attributes.length) {
        for (f.i = 0; f.i < f.element.attributes.length; f.i++) {
          if (attributeAllowed(nodeName, f.element.attributes[f.i].name) && !(f.element.nodeName == "VIDEO" && f.element.attributes[f.i].name == "src")) {
            f.string += " " + transformAttribute(f.element, f.element.attributes[f.i]);
          }
        }
      }

      if (keepStyle && nodeName != "map") {
        // ENML doesn't allow style for map
        f.string += style.style;
      } else if (f.element.nodeName == "HIGHLIGHT") {
        f.string += " style=\"x-evernote:highlighted;background-color:#F6EE96;\"";
      }
      f.string += ">";

      if (keepStyle) {
        if (style.before) {
          f.string += style.before;
        }
      }

      f.i = 0;
    }

    while (f.i < f.element.childNodes.length) {
      if (f.element.childNodes[f.i].nodeType == Node.TEXT_NODE) {
        var text;

        if (range && f.element.childNodes[f.i] === range.startContainer) {
          text = escapeHTML(f.element.childNodes[f.i].textContent.substr(range.startOffset));
        }
        else if (range && f.element.childNodes[f.i] === range.endContainer) {
          text = escapeHTML(f.element.childNodes[f.i].textContent.substr(0, range.endOffset));
        }
        else if (range && !range.intersectsNode(f.element.childNodes[f.i])) {
          text = "";
        }
        else {
          text = escapeHTML(f.element.childNodes[f.i].textContent);
        }
        // text = text.replace(/\s+/g, " "); // @TODO: Enable in production.
        f.string += text;
        f.i++;
      }
      else if(f.element.childNodes[f.i].nodeType == Node.ELEMENT_NODE && childrenAllowed(f.element)) {
        stack.push({element: f.element.childNodes[f.i], string: f.string, i: 0, after: null});
        f.i++;
        recurse();
        return;
      }
      else {
        f.i++;
      }
    }

    if (keepStyle) {
      if (f.after) {
        f.string += f.after;
      }
    }

    f.string += "</" + transformNode(f.element) + ">";

    stack.pop();
    if (stack.length) {
      stack[stack.length - 1].string = f.string;
      recurse();
    }
    else {
      doneRecursing(f.string);
    }
  }

  function childrenAllowed(el) {
    if (el.nodeName == "VIDEO") {
      return false;
    }
    return true;
  }

  // If someone specifies "background: 0" in CSS, chrome expands that out to:
  // background-position: 0px 50%; background-repeat: initial initial;
  // We restore the original here.
  function rebackgroudifyCss(map) {
    if (map["background-position"] && map["background-repeat"]) {
      if (map["background-position"].trim() == "0px 50%" && map["background-repeat"].trim() == "initial initial") {
        for (var prop in map) {
          if (prop.match(/background/)) {
            delete map[prop];
          }
        }
        map["background"] = "0";
      }
    }
  }
  // @TODO: this is almost certainly horribly incomplete. NOTE: Actually it works remarkably well.
  function parseCssText(str) {
    var val = {};
    // regex handles the case where a data uri is used in css
    var props = str.split(/;(?!.[^\(]+?\))\s*/);
    for (var i = 0; i < props.length; i++) {
      props[i] = props[i].trim();
      if (props[i]) {
        var splitIdx = props[i].indexOf(":");
        var name = props[i].substr(0, splitIdx).trim();
        var value = props[i].substr(splitIdx + 1).trim();
        if (name && value) {
          // could have multiple urls in background-image, so process each one
          var regex = new RegExp("(url\\(.+?\\))", "g");
          var match = regex.exec(value);
          var valueCopy = value;
          while (match) {
            var orig = match[1];
            var updated = orig;
            // wikipedia mysteriously has a background-image url like
            // http://bits.wikimeida.org/en.wikimedia.org/data:image/png...
            // so correct it to just the data URI
            if (/url\(.+?(data:.+?;.+?,.+?)\)/.test(updated)) {
              updated = "url(" + /url\(.+?(data:.+?;.+?,.+?)\)/.exec(updated)[1] + ")";
            }
            // and sometimes a url will be in the form of "//upload.wikimedia.org"
            // which the desktop clients don't know how to show, so add an http
            if (/url\(\/\/(.+?)\)/.test(updated)) {
              updated = "url(http://" + /url\(\/\/(.+?)\)/.exec(updated)[1] + ")";
            }
            // escape chars in urls
            if (/url\((.+?)\)/.test(updated)) {
              updated = "url(" + GlobalUtils.escapeXML(GlobalUtils.unescapeXML(/url\((.+?)\)/.exec(updated)[1])) + ")";
            }
            valueCopy = valueCopy.replace(orig, updated);
            match = regex.exec(value);
          }
          value = valueCopy;
          val[name.toLowerCase()] = value;
        }
      }
    }
    rebackgroudifyCss(val);
    return val;
  }

  function objectifyCssRule(rule) {
    var styleMap = {};
    if (rule.style.cssText) {
      var styles;
      if (!rule.style.savedCssObj) {
        rule.style.savedCssObj = parseCssText(rule.style.cssText);
      }
      styles = rule.style.savedCssObj;

      for (var k in styles) {
        styleMap[k] = styles[k];
      }
    }
    return styleMap;
  }

  function specifyImgDims(el) {
    if (el.nodeName.toLowerCase() == "img") {
      if (!el.attributes.width) {
        el.setAttribute("width", el.width);
      }
      if (!el.attributes.height) {
        el.setAttribute("height", el.height);
      }
    }
  }

  // some HTML attributes were deprecated in favor of CSS attributes. convert
  // them here so that we retain these style elements. e.g. background
  function convertHTMLAttributesToCSSAttributes(el) {
    if (el.style) {
      // TODO alink, link, vlink
      if (el.hasAttribute("background")) {
        el.style.backgroundImage = "url(" + el.getAttribute("background") + ")";
      }
      if (el.hasAttribute("bgcolor")) {
        el.style.backgroundColor = el.getAttribute("bgcolor");
      }
      if (el.hasAttribute("text")) {
        el.style.color = el.getAttribute("text");
      }
    }
  }

  // @TODO: This is wrong for comma-separated selectors. We should split them and check each one individually.
  function getSelectorSpecificity(sel) {
    var matchers = {
      "ids": {
        "regex": /#[A-Z]+/ig,
        "count": 0
      },
      "classes": {
        "regex": /\.[A-Z]+/ig,
        "count": 0
      },
      "attrs": {
        "regex": /\[.*?\]/g,
        "count": 0
      },
      "pseudos": {
        "regex": /:+[A-Z]+/ig,
        "count": 0
      },
      "pseudoEls": {
        "regex": /:+(first-line|first-letter|before|after)/ig,
        "count": 0
      },
      "types": {
        "regex": /(^|\s)[A-Z]+/ig,
        "count": 0
      }
    };

    for (var i in matchers) {
      var re = matchers[i].regex;
      while (re.exec(sel)) {
        matchers[i].count++;
      }
    }

    matchers.pseudoClasses = {};
    matchers.pseudoClasses.count = matchers.pseudos.count - matchers.pseudoEls.count;

    var first = matchers.ids.count;
    var second = matchers.classes.count + matchers.attrs.count + matchers.pseudoClasses.count;
    var third = matchers.types.count + matchers.pseudoEls.count;

    var score = (first * 256 * 256) + (second * 256) + third;

    return score;
  }

  function splitSelectorList(sel) {
    var sels = [];
    var lastStart = 0;
    var i = 0;
    var quoted = "";
    for (i = 0; i < sel.length; i++) {
      if (!quoted) {
        if (sel[i] == "'" || sel[i] == "\"") {
          quoted = sel[i];
        }
        else if (sel[i] == ",") {
          sels.push(sel.substring(lastStart, i).trim());
          lastStart = i + 1;
        }
      }
      else {
        if (sel[i] == quoted) {
          quoted = "";
        }
      }
    }
    sels.push(sel.substr(lastStart).trim());
    return sels;
  }

  // Properties listed here http://www.w3.org/TR/CSS21/propidx.html that are both 'visual' and inherited. This is used
  // to handle inheritance into generated content blocks properly. Note that this is a list from CSS 2.1 and may miss
  // items that were added in CSS3.
  var inheritableCSSProperties = {
    'border-collapse': true,
    'border-spacing': true,
    'caption-side': true,
    'color': true,
    'cursor': true,
    'direction': true,
    'empty-cells': true,
    'font-family': true,
    'font-size': true,
    'font-style': true,
    'font-variant': true,
    'font-weight': true,
    'font': true,
    'letter-spacing': true,
    'line-height': true,
    'list-style-image': true,
    'list-style-position': true,
    'list-style-type': true,
    'list-style': true,
    'orphans': true,
    'quotes': true,
    'text-align': true,
    'text-indent': true,
    'text-transform': true,
    'visibility': true,
    'white-space': true,
    'widows': true,
    'word-spacing': true
  };
  function cssPropertyIsInheritable(property) {
    if (inheritableCSSProperties[property.toLowerCase()]) {
      return true;
    }
    return false;
  }

  function fixQuirksModeTableInheritance(el, map) {
    var nodeName = el.nodeName.toLowerCase();
    if (nodeName == "table" || nodeName == "caption") {
      if (_window.document.compatMode == "CSS1Compat") {
        // console.log("standards mode document, forcing table inheritance.");
        map["font-size"] = {value: "inherit", score: 0};
        map["font-weight"] = {value: "inherit", score: 0};
        map["font-style"] = {value: "inherit", score: 0};
        map["font-variant"] = {value: "inherit", score: 0};
      }
    }
  }

  // @TODO: there are almost certainly more to add here.
  function clearOverridden(name, map) {
    if (name == "padding") {
      delete map["padding-top"];
      delete map["padding-bottom"];
      delete map["padding-left"];
      delete map["padding-right"];
    } else if (name == "margin") {
      delete map["margin-top"];
      delete map["margin-bottom"];
      delete map["margin-left"];
      delete map["margin-right"];
    } else if (name == "background") {
      delete map["background-color"];
      delete map["background-position"];
      delete map["background-size"];
      delete map["background-repeat"];
      delete map["background-origin"];
      delete map["background-clip"];
      delete map["background-attachment"];
      delete map["background-image"];
      clearOverridden("background-position", map);
      clearOverridden("background-repeat", map);
    } else if (name == "background-position") {
      delete map["background-position-x"];
      delete map["background-position-y"];
    } else if (name == "background-repeat") {
      delete map["background-repeat-x"];
      delete map["background-repeat-y"];
    } else if (name == "border") {
      delete map["border-width"];
      delete map["border-style"];
      delete map["border-color"];
      clearOverridden("border-top", map);
      clearOverridden("border-right", map);
      clearOverridden("border-bottom", map);
      clearOverridden("border-left", map);
    } else if (name == "border-top") {
      delete map["border-top-width"];
      delete map["border-top-style"];
      delete map["border-top-color"];
    } else if (name == "border-right") {
      delete map["border-right-width"];
      delete map["border-right-style"];
      delete map["border-right-color"];
    } else if (name == "border-bottom") {
      delete map["border-bottom-width"];
      delete map["border-bottom-style"];
      delete map["border-bottom-color"];
    } else if (name == "border-left") {
      delete map["border-left-width"];
      delete map["border-left-style"];
      delete map["border-left-color"];
    }
  }

  // https://stackoverflow.com/questions/2952667/find-all-css-rules-that-apply-to-an-element
  // Calling getCss(document.getElementById('elementId')) will return an array with an element for each CSS rule that matches the passed element.
  // Currently a workaround for the removal of getMatchedCSSRules since Chrome 63
  function getCss(el) {
      var sheets = document.styleSheets, ret = [];
      el.matches = el.matches || el.webkitMatchesSelector || el.mozMatchesSelector
          || el.msMatchesSelector || el.oMatchesSelector;
      for (var i in sheets) {
          try {
              var rules = sheets[i].rules || sheets[i].cssRules;
              for (var r in rules) {
                  if (el.matches(rules[r].selectorText)) {
                      ret.push(rules[r]);
                  }
              }
          } catch (exception) {
//            console.log("css(): " + exception);
          }
      }
      return ret;
  }

  // ================================================================================================================
  // And the following block does "fast" CSS resolution, where we can call getMatchedCSSRules, which is currently (and
  // maybe always will be) Webkit-only.
  function resolveStyle(el, stripStyleList) {
    try {
      convertHTMLAttributesToCSSAttributes(el);

      var style = "";
      var originalStyle = null;
      var before = {};
      var after = {};
      if (el.attributes && el.attributes.style) {
        originalStyle = parseCssText(el.attributes.style.value);
      }

      var styleMap = {};

      fixQuirksModeTableInheritance(el, styleMap);

      var rules = _window.getMatchedCSSRules ? _window.getMatchedCSSRules(el) : getCss(el); // getMatchedCSSRules is removed in Chromium 63

      if (rules && rules.length) {
        for (var i = 0; i < rules.length; i++) {

          var specScore = 0;
          var ignoreVisited = false;
          if (rules[i].selectorText.match(/:visited/i)) {
            ignoreVisited = true;
          }

          var selectors = splitSelectorList(rules[i].selectorText);
          for (var j = 0; j < selectors.length; j++) {
            var matches;
            try {
              matches = el.webkitMatchesSelector(selectors[j]);
            }
            catch (e){
              console.warn("Couldn't match against selector " + selectors[j] + " in: " + rules[i].selectorText);
            }
            if (matches) {
              ignoreVisited = false;
              // If we match anything, then we'll ignore whether we matched a
              // visited rule, as matchesSelector will always return 'false' for
              // these rules.
              var possibleSpec = getSelectorSpecificity(selectors[j]);
              if (possibleSpec >= specScore) {
                specScore = possibleSpec;
              }
            }
          }

          if (ignoreVisited) {
            // console.log("should ignore visited selector: " + rules[i].selectorText);
            continue;
          }

          var ruleObj = objectifyCssRule(rules[i]);

          // if it's a counter-increment/reset, keep track of that in javascript
          // and don't include that in the css
          if (ruleObj["counter-reset"]) {
            var counterResetParts = ruleObj["counter-reset"].split(/\s+/);
            if (counterResetParts.length > 0) {
              var crp = 0;
              while (crp < counterResetParts.length) {
                if (/^\d+$/.test(counterResetParts[crp + 1])) {
                  counters[counterResetParts[crp]] = parseInt(counterResetParts[crp + 1]);
                  crp += 2;
                } else {
                  counters[counterResetParts[crp++]] = 0;
                }
              }
            }
          }
          if (ruleObj["counter-increment"]) {
            var counterIncParts = ruleObj["counter-increment"].split(/\s+/);
            if (counterIncParts.length > 0) {
              var cip = 0;
              while (cip < counterIncParts.length) {
                if (!counters[counterIncParts[cip]]) {
                  counters[counterIncParts[cip]] = 0;
                }
                if (/^\d+$/.test(counterIncParts[cip + 1])) {
                  counters[counterIncParts[cip]] += parseInt(counterIncParts[cip + 1]);
                  cip += 2;
                } else {
                  counters[counterIncParts[cip++]]++;
                }
              }
            }
          }

          for (var k in ruleObj) {

            // Skip invalid properties.
            var jsPropName = k.replace(/^-/, "").replace(/-[a-z]/g, function(str){ return str[1].toUpperCase(); });
            if (!rules[i].style[jsPropName] || jsPropName === "counterIncrement" || jsPropName === "counterReset") {
              continue;
            }

            // See if there was a pre=existing score for this property.
            var oldScore = 0;
            if (styleMap[k]) {
              oldScore = styleMap[k].score;
            }

            // Adjusts scores for !important rules.
            var ruleScore = specScore;
            if (ruleObj[k].match(/!important\s*$/i)) {
              ruleScore += (256 * 256 * 256);
              ruleObj[k] = ruleObj[k].replace(/\s*!important\s*$/i, "");
            }
            // Replace if greater or equal.
            if (ruleScore >= oldScore) {
              clearOverridden(k, styleMap);
              styleMap[k] = {value: ruleObj[k], score: ruleScore};
            }
          }
        }
      }

      // For handling "before" and "after" pseudo-elements.
      for (var i = 0; i < pseudoElementRules.length; i++) {
        var rule = pseudoElementRules[i];
        var match;
        if (el.webkitMatchesSelector) {
          // Because we don't have a real CSS engine, sometimes the following heuristic regex replacement leaves us with
          // invalid CSS selectors, which will throw "DOM Exception 12" errors when we try and match against them, so we
          // just catch that and ignore it to keep ourselves from failing out entirely.
          try {
            match = el.webkitMatchesSelector(rule.selectorText.replace(/(:?:before)|(:?:after)/g, ""));
          }
          catch (e) {
            // DOM Exception 12.
          }
        }
        if (match) {

          var matchBefore = false;
          var matchAfter = false;
          if (rule.selectorText.match(/:?:before/)) {
            matchBefore = true;
          }
          if (rule.selectorText.match(/:?:after/)) {
            matchAfter = true;
          }

          // Inherit as per: http://www.w3.org/TR/CSS21/generate.html;
          // @TODO: Implement specificity here.
          for (var k in styleMap) {
            if (matchBefore && cssPropertyIsInheritable(k)) {
              before[k] = styleMap[k].value;
            }
            if (matchAfter && cssPropertyIsInheritable(k)) {
              after[k] = styleMap[k].value;
            }
          }
          var generated = objectifyCssRule(rule);
          for (var k in generated) {
            if (matchBefore) {
              before[k] = generated[k];
            }
            if (matchAfter) {
              after[k] = generated[k];
            }
          }
        }
      }

      var sections = [before, after];
      for (var entry = 0; entry < sections.length; entry++) {
        var map = sections[entry];
        var pseudoStyle = "";
        var content = "";
        var count = 0;
        // process counter-reset and counter-increment before content counter
        // according to http://www.w3.org/TR/CSS21/generate.html#counters
        if (map["counter-reset"]) {
          var counterResetParts = map["counter-reset"].split(/\s+/);
          var crp = 0;
          while (crp < counterResetParts.length) {
            if (/^\d+$/.test(counterResetParts[crp + 1])) {
              counters[counterResetParts[crp]] = parseInt(counterResetParts[crp + 1]);
              crp += 2;
            } else {
              counters[counterResetParts[crp++]] = 0;
            }
          }
        }
        if (map["counter-increment"]) {
          var counterIncParts = map["counter-increment"].split(/\s+/);
          var cip = 0;
          while (cip < counterIncParts.length) {
            if (!counters[counterIncParts[cip]]) {
              counters[counterIncParts[cip]] = 0;
            }
            if (/^\d+$/.test(counterIncParts[cip + 1])) {
              counters[counterIncParts[cip]] += parseInt(counterIncParts[cip + 1]);
              cip += 2;
            } else {
              counters[counterIncParts[cip++]]++;
            }
          }
        }
        for (var j in map) {
          if (j == "content") {
            // Ghetto-parses quoted strings, mostly.
            var content = map[j];
            content = content.trim();
            content = content.replace(/\s+!important$/, "");
            if (content == "none") content = "\"\"";
            if (content.match(/^'/)) {
              content = content.replace(/^'(.*?)'.*/, "$1");
            }
            else if (content.match(/^"/)) {
              content = content.replace(/^"(.*?)".*/, "$1");
            }

            if (content.match(/^url\((.*)\)/)) {
              var contentUrl = content.match(/^url\((.*)\)/)[1];
              content = "<img src='" + contentUrl + "'></img>";
            } else if (content.match(/^\s*counter\(.*?\)/)) {
              // sometimes this is a comma-separated list, but that's not standard
              // so just take the first one in the list, like the browser does
              var counterName = content.match(/^\s*counter\((.*?)\)/)[1].split(",")[0];
              if (typeof counters[counterName] === "number") {
                content = GlobalUtils.escapeXML("" + counters[counterName]);
              }
              // TODO(cchan): handle counters (not just counter) too
            } else if (content.match(/^\s*attr\((.*?)\)/)) {
              var attrName = content.match(/^\s*attr\((.*?)\)/)[1];
              if (el.getAttribute(attrName)) {
                content = GlobalUtils.escapeXML("" + el.getAttribute(attrName));
              } else {
                content = "";
              }
            } else if (content.match(/^\s*-webkit-image-set\(.+\)/)) {
              pseudoStyle += "background:" + content + ";";
              content = "";
            }
          } else if (j !== "counter-reset" && j !== "counter-increment"){
            pseudoStyle += j + ":" + map[j] + ";";
          }
          count++;
        }
        if (count) {
          pseudoStyle = "<span style=\"" + pseudoStyle + "\">" + content + "</span>";
          if (entry == 0) {
            before = pseudoStyle;
          }
          else {
            after = pseudoStyle;
          }
        }
      }

      // Clear out anything tyhat didn't get set.
      if (typeof before != "string") before = null;
      if (typeof after != "string") after = null;

      // el == element is a special hack to remove padding and such from the main element, so that it doesn't get
      // positioned strangely in our note view.
      if (el == element) {
        stripStyleList = strippableProperties;
      }
      if (stripStyleList) {
        for (var i = 0; i < stripStyleList.length; i++) {
          if (styleMap[stripStyleList[i]]) {
            delete styleMap[stripStyleList[i]];
          }
        }
      }

      // Hacks and adjustments.
      // ================================================================================
      fixHeight(el, styleMap);
      fixUserSelect(styleMap);
      // ================================================================================

      if (originalStyle) {
        var matcher = /url\(['"]?(.*?)['"]?\)/i;
        for (var k in originalStyle) {

          // If our style tag contains attrbiutes with URLs, we may need to canonicalize them,
          if (matcher.test(originalStyle[k])) {
            var url = matcher.exec(originalStyle[k])[1];
            var linkBase = _window.document.location.href;
            if (base && base.href) {
              linkBase = base.href;
            }
            var fixed = reconstituteUrl(linkBase, url);
            originalStyle[k] = originalStyle[k].replace("url(\"" + url + "\")", fixed)
              .replace("url('" + url + "')", fixed).replace("url(" + url + ")", fixed);
          }

          styleMap[k] = {value: originalStyle[k]};
        }
      }

      for (var i in styleMap) {
        style += i + ":" + styleMap[i].value + ";";
      }

      if (style) {
        style = " style=\"" + escapeAttr(style) + "\"";
      }
      else {
        style = "";
      }
    } catch (exception) {
      return {style: null, before: null, after: null, map: null};
    }

    return {style: style, before: before, after: after, map: styleMap};
  }

  // ================================================================================================================


  // Height percentage adjustment to auto, as per here:
  // https://developer.mozilla.org/en/CSS/height
  function fixHeight(el, styleMap) {
    if (styleMap["height"] && styleMap["height"].value.match(/%$/)) {
      var fixed = styleMap["height"].value;
      try {
        if (_window.getComputedStyle(el).position != "fixed") {
          // var containingHeight = el.parentNode.style.height;
          var containingHeight = _window.getComputedStyle(el.parentNode).height;
          if (!containingHeight) {
            fixed = "auto";
          }
        }
      }
      catch(e) { /*in case we can't resolve the parent's style chain. */ }
      styleMap["height"].value = fixed;
    }
  }

  // This keeps notes modifiable.
  function fixUserSelect(styleMap) {
    var selectProps = [
      "-webkit-user-select",
      "-moz-user-select",
      "-ms-user-select",
      "user-select",
      "-webkit-user-modify",
      "-moz-user-modify",
      "-ms-user-modify",
      "user-modify"
    ];
    for (var i = 0; i < selectProps; i++) {
      if (styleMap[selectProps[i]]) {
        delete styleMap[selectProps[i]];
      }
    }
  }

  function transformAttribute(element, attrObj) {
    var val = null;
    if (attrObj.name.toLowerCase() == "href") {
      // The following magic returns a fully-qualified path for 'href' instead of whatever the DOM says in the 'href'
      // attribute. could be a relative url if it's href.animVal
      var href = element.href.animVal ? element.href.animVal : element.href;
      if (href && href.match(/^javascript/i)) {
        href = "#";
      } else if (!/^https?:\/\//.test(href)) {
        href = element.baseURI + href;
      }
      val = escapeAttr(href);
    }
    else if (attrObj.name.toLowerCase() == "src") {
      // <img src /> returns the page's url for element.src which is wrong,
      // so make it blank
      var src = element.src;
      if (attrObj.value.trim() == "") {
        src = "";
      }
      if (src && src.match(/^javascript/i)) src = "#";
      src = convertImgSrcToBase64IfPossible(element, src);
      val = escapeAttr(src);
    }

    if (val === null) {
      // strip newlines from attributes only, not the rest of the HTML
      val = escapeAttr(attrObj.value.replace(/\u000A/g, ""));
    }

    return attrObj.name + "=\"" + val + "\"";
  }

  // convert the images with http/https urls to dataurl form. it's faster
  // to extract them from the <img> elements here than in the background.
  // this only works for images that have the same domain as the web page.
  // the ones without the same domain will be handled with an XHR in the
  // background which is faster than the img/canvas approach used here.
  function convertImgSrcToBase64IfPossible(element, src, naturalSize) {
    if (!element) {
      return src;
    }
    var win = _window || window;
    if (element.nodeName == "IMG" && /^https?:\/\//i.test(src) && !/\.gif($|\?)/i.test(src)
        && (element.naturalWidth || element.naturalHeight)) {
      // convert same-domain images to base64
      var imgOrigin = /^(https?:\/\/.[^\/]+)\/?/.exec(src)[1];
      if (imgOrigin === win.document.location.origin) {
        var canvas = win.document.createElement("canvas");
        canvas.width = (naturalSize ? element.naturalWidth : element.width) || 1;
        canvas.height = (naturalSize ? element.naturalHeight : element.height) || 1;
        canvas.getContext("2d").drawImage(element, 0, 0, (naturalSize ? element.naturalWidth : element.width), (naturalSize ? element.naturalHeight : element.height));
        try {
          if (/\.jpe?g$/.test(src)) {
            return canvas.toDataURL("image/jpeg");
          } else {
            return canvas.toDataURL("image/png");
          }
        } catch(err) {
          // need to try-catch this because even though an image passed the
          // cross-origin check, it could still be redirect to something
          // that's out of this domain. 18 is the right securityexception
          if (err.code != 18) {
            throw err;
          }
          return src;
        } finally {
          canvas = null;
        }
      } else {
        return src;
      }
    } else {
      return src;
    }
  }

  function restoreAncestors() {
    var front = "";
    var back = "";
    var current = element.parentNode;
    while (current && current.parentNode) {
      var style = resolveStyle(current, strippableProperties).style;
      var type = transformNode(current);

      if (nodeAllowed(type)) {
        front = "<" + type + style + ">" + front;
        back = back + "</" + type + ">";
      }
      current = current.parentNode;
    }

    // Last one, set the base font size
    front = "<div style=\"font-size: 16px\">" + front;
    back = back + "</div>";

    return {front: front, back: back};
  }

  function doneRecursing(str, error) {
    blocked = false;

    if (!error) {
      var ancestors = {front: "", back: ""};
      if (keepStyle) {
        ancestors = restoreAncestors();
      }
      str = ancestors.front + str + ancestors.back;
    }

    for (var i = 0; i < stylesToRemove.length; i++) {
      stylesToRemove[i].parentNode.removeChild(stylesToRemove[i]);
    }
    stylesToRemove = [];

    for (var i = 0; i < callbacks.length; i++) {
      try {
        callbacks[i](str, error);
      }
      catch (e) {
        console.warn("Couldn't run 'serialize' callback: " + e.stack || e.trace);
      }
    }
    callbacks = [];
  }

  function escapeAttr(str) {
    if (!str) return "";
    return GlobalUtils.escapeXML(str);
  }

  this.convertImgSrcToBase64IfPossible = convertImgSrcToBase64IfPossible;
  this.serialize = serialize;

  Object.preventExtensions(this);
}

Object.preventExtensions(HtmlSerializer);